from datetime import timedelta
from typing import Any

from PySubtrans.Substitutions import Substitutions
from PySubtrans.TranslationPrompt import TranslationPrompt
from PySubtrans.SubtitleError import SubtitleError
from PySubtrans.Helpers.SubtitleHelpers import AddOrUpdateLine, MergeSubtitles
from PySubtrans.SubtitleLine import SubtitleLine
from PySubtrans.Translation import Translation

class SubtitleBatch:
    def __init__(self, dct : dict[str,Any]|None = None):
        dct = dct or {}
        self.scene : int = dct.get('scene', 0)
        self.number : int = dct.get('batch') or dct.get('number') or 0
        self.summary : str|None = dct.get('summary')
        self.context : dict[str,str|list|dict] = dct.get('context', {})
        self.errors : list[str|SubtitleError] = dct.get('errors', [])
        self._originals : list[SubtitleLine] = dct.get('originals', []) or dct.get('subtitles', [])
        self._translated : list[SubtitleLine] = dct.get('translated', [])
        self.translation : Translation|None = dct.get('translation')
        self.prompt : TranslationPrompt|None = dct.get('prompt')

    def __str__(self) -> str:
        return f"SubtitleBatch: {str(self.number)} in scene {str(self.scene)} with {self.size} lines"

    def __repr__(self) -> str:
        return str(self)

    @property
    def originals(self) -> list[SubtitleLine]:
        """ Get the list of original lines in the batch """
        return self._originals

    @property
    def size(self) -> int:
        """ Get the number of original lines in the batch """
        return len(self._originals)

    @property
    def translated(self) -> list[SubtitleLine]:
        """ Get the list of translated lines in the batch """
        return self._translated

    @property
    def untranslated(self) -> list[SubtitleLine]:
        """ Get the list of lines in the batch which have no translation """
        return [sub for sub in self.originals if not sub.translated]

    @property
    def all_translated(self) -> bool:
        """ Check if all original lines have a translation """
        return True if (self.translated and (len(self.translated) == len(self.originals))) else False

    @property
    def any_translated(self) -> bool:
        """ Check if any original lines have a translation """
        return len(self.translated or []) > 0

    @property
    def start(self) -> timedelta|None:
        """ Get the start time of the first line in the batch """
        return self.originals[0].start if self.originals else None

    @property
    def srt_start(self) -> str:
        """ Get the start time of the first line in SRT format """
        return self.originals[0].srt_start if self.originals else ""

    @property
    def txt_start(self) -> str:
        """ Get the start time of the first line in human-readable format """
        return self.originals[0].txt_start if self.originals else ""

    @property
    def end(self) -> timedelta|None:
        """ Get the end time of the last line in the batch """
        return self.originals[-1].end if self.originals else None

    @property
    def srt_end(self) -> str:
        """ Get the end time of the last line in SRT format """
        return self.originals[-1].srt_end if self.originals else ""

    @property
    def txt_end(self) -> str:
        """ Get the end time of the last line in human-readable format """
        return self.originals[-1].txt_end if self.originals else ""

    @property
    def duration(self) -> timedelta:
        """ Get the time delta between the start and end of the batch """
        return self.end - self.start if self.start is not None and self.end else timedelta(seconds=0)

    @property
    def first_line_number(self) -> int|None:
        """ Get the first line number in the batch """
        return self.originals[0].number if self.originals else None

    @property
    def last_line_number(self) -> int|None:
        """ Get the last line number in the batch """
        return self.originals[-1].number if self.originals else None

    @property
    def error_messages(self) -> list[str]:
        """ Get the list of error messages associated with the batch """
        return [str(error) for error in self.errors if isinstance(error, (str, Exception))]

    @originals.setter
    def originals(self, value : list[SubtitleLine]|list[str]):
        """
        Update the original text for the batch
        """
        lines = [SubtitleLine(line) for line in value] if value else []
        self._originals = [line for line in lines if line.number]

    @translated.setter
    def translated(self, value : list[SubtitleLine]|list[str]):
        """ 
        Update the translated text for the batch
        """
        lines = [SubtitleLine(line) for line in value] if value else []
        self._translated = [line for line in lines if line.number]

    def AddLine(self, line : SubtitleLine|str):
        """
        Insert a line into the batch or replace an existing line
        """
        AddOrUpdateLine(self._originals, SubtitleLine(line))

    def AddTranslatedLine(self, line : SubtitleLine|str):
       """ Insert a translated line into the batch or replace an existing translation """
       AddOrUpdateLine(self._translated, SubtitleLine(line))

    def HasTranslatedLine(self, line_number : int) -> bool:
        """ Check if the batch has a translated line with the given number """
        if not self.first_line_number or not self.last_line_number:
            return False
        if line_number < self.first_line_number or line_number > self.last_line_number:
            return False

        return any(line for line in self._translated if line.number == line_number)

    def GetOriginalLine(self, line_number : int) -> SubtitleLine|None:
        """ Get an original line from the batch by its number """
        return next((line for line in self._originals if line.number == line_number), None)

    def GetTranslatedLine(self, line_number : int) -> SubtitleLine|None:
        """ Get a translated line from the batch by its number """
        return next((line for line in self._translated if line.number == line_number), None)

    def AddContext(self, key : str, value : str|list[str]|dict[str,Any]):
        self.context[key] = value

    def GetContext(self, key : str) -> str|dict|list|None:
        return self.context.get(key)

    def SetContext(self, context : dict[str,Any]):
        self.context = context.copy()

    def UpdateContext(self, update : dict[str,Any]) -> bool:
        if not self.context:
            self.context = {}

        updated = False
        for key in update.keys():
            if key == 'summary':
                if update[key] != self.summary:
                    self.summary = update[key]
                    updated = True

            elif update[key] != self.context.get(key):
                self.context[key] = update[key]
                updated = True

        return updated

    def PerformInputSubstitutions(self, substitutions : Substitutions) -> dict[str,str]|None:
        """
        Perform any word/phrase substitutions on source text
        """
        if substitutions and self.originals:
            lines : list[str] = [item.text for item in self.originals if item.text]

            lines, replacements = substitutions.PerformSubstitutionsOnAll(lines)

            if replacements:
                self.AddContext('input_replacements', replacements)
                for item in self.originals:
                    if item.text:
                        # Replace the text in the original lines
                        item.text = replacements.get(item.text, item.text)

            return replacements

    def PerformOutputSubstitutions(self, substitutions : Substitutions) -> dict[str,str]|None:
        """
        Perform any word/phrase substitutions on translated text
        """
        if substitutions and self.translated:
            lines = [item.text for item in self.translated if item.text]

            _, replacements = substitutions.PerformSubstitutionsOnAll(lines)

            if replacements:
                self.AddContext('output_replacements', replacements)
                for item in self.translated:
                    if item.text:
                        item.text = replacements.get(item.text) or item.text

            return replacements

    def MergeLines(self, line_numbers : list[int]) -> tuple[SubtitleLine, SubtitleLine|None]:
        """
        Merge multiple lines into a single line with the same start/end times
        """
        lines = [line for line in self.originals if line.number in line_numbers]

        if len(lines) < 2:
            raise SubtitleError(f"Cannot merge {len(lines)} lines")

        first_index = self.originals.index(lines[0])
        last_index = self.originals.index(lines[-1])

        merged = MergeSubtitles(lines)
        self._originals = self.originals[:first_index] + [ merged ] + self.originals[last_index + 1:]

        translated_lines = [line for line in self.translated if line.number in line_numbers]

        if translated_lines and len(translated_lines) > 1:
            first_translated_index = self.translated.index(translated_lines[0])
            last_translated_index = self.translated.index(translated_lines[-1])
            merged_translated = MergeSubtitles(translated_lines)
            self._translated = self.translated[:first_translated_index] + [ merged_translated ] + self.translated[last_translated_index + 1:]

            return merged, merged_translated

        return merged, None

    def DeleteLines(self, line_numbers : list[int]) -> tuple[list[SubtitleLine], list[SubtitleLine]]:
        """
        Delete lines from the batch
        """
        originals = [line for line in self.originals if line.number not in line_numbers]
        translated = [line for line in self.translated if line.number not in line_numbers]

        if len(originals) == len(self.originals) and len(translated) == len(self.translated):
            return [],[]

        deleted_originals = [line for line in self.originals if line.number in line_numbers]
        deleted_translated = [line for line in self.translated if line.number in line_numbers]

        self._originals = originals
        self._translated = translated

        return deleted_originals, deleted_translated

    def InsertOriginalLine(self, line : SubtitleLine):
        """
        Insert a line into the batch
        """
        if not line:
            raise SubtitleError("No line provided to insert")

        if not self.originals:
            self.originals = [line]

        if self.first_line_number is not None and line.number < self.first_line_number:
            self.originals.insert(0, line)

        elif self.last_line_number is not None and line.number > self.last_line_number:
            self.originals.append(line)

        else:
            for index, item in enumerate(self.originals):
                if item.number >= line.number:
                    self.originals.insert(index, line)
                    break

    def InsertTranslatedLine(self, line : SubtitleLine):
        """
        Insert a translated line into the batch
        """
        if not line:
            raise SubtitleError("No line provided to insert")

        if not self.translated:
            self.translated = [line]

        if line.number < self.translated[0].number:
            self.translated.insert(0, line)

        elif line.number > self.translated[-1].number:
            self.translated.append(line)

        else:
            for index, item in enumerate(self.translated):
                if item.number >= line.number:
                    self.translated.insert(index, line)
                    break

    def InsertLines(self, originals: list[SubtitleLine], translated: list[SubtitleLine]|None = None):
        """
        Insert multiple lines into the batch, with optional translations
        """
        if not originals:
            raise SubtitleError("No original lines provided to insert")

        originals = sorted(originals, key=lambda item: item.number)

        for line in originals:
            self.InsertOriginalLine(line)

        if translated:
            sorted_lines = sorted(translated, key=lambda item: item.number)
            for line in sorted_lines:
                self.InsertTranslatedLine(line)




def VersionNumberLessThan(version1: str, version2: str) -> bool:
    """Compare two version strings and return True if version1 is less than version2"""
    if not version1:
        return True
        
    # Strip 'v' prefix if present
    v1 = version1[1:] if version1.startswith('v') else version1
    v2 = version2[1:] if version2.startswith('v') else version2
    
    # Split into components and convert to integers
    v1_parts = [int(x) for x in v1.split('.')]
    v2_parts = [int(x) for x in v2.split('.')]
    
    # Compare each component
    for i in range(max(len(v1_parts), len(v2_parts))):
        v1_part = v1_parts[i] if i < len(v1_parts) else 0
        v2_part = v2_parts[i] if i < len(v2_parts) else 0
        if v1_part < v2_part:
            return True
        if v1_part > v2_part:
            return False

    return False



import importlib.util
import logging
import os

from PySubtrans.Options import SettingsType, env_float
from PySubtrans.SettingsType import GuiSettingsType, SettingsType

if not importlib.util.find_spec("openai"):
    from PySubtrans.Helpers.Localization import _
    logging.debug(_("OpenAI SDK is not installed. OpenAI provider will not be available"))
else:
    try:
        import openai   # type: ignore

        from PySubtrans.Helpers.Localization import _
        from PySubtrans.Providers.Clients.ChatGPTClient import ChatGPTClient
        from PySubtrans.Providers.Clients.OpenAIReasoningClient import OpenAIReasoningClient
        from PySubtrans.SubtitleError import ProviderError
        from PySubtrans.TranslationClient import TranslationClient
        from PySubtrans.TranslationProvider import TranslationProvider

        class OpenAiProvider(TranslationProvider):
            name = "OpenAI"

            information = """
            <p>Select the AI <a href="https://platform.openai.com/docs/models">model</a> to use as a translator.</p>
            <p>Note that different models have different <a href="https://openai.com/pricing">costs</a> and limitations.</p>
            In particular, the number of tokens supported by each model will affect the batch size that it can handle.
            This will depend on the contents but as a rule of thumb 16K token models can handle 80-100 lines per batch.</p>
            <p>GPT4 models are substantially more expensive but are better at following instructions,
            e.g. using provided names, and may be better for more complex translations and less common languages.</p>
            """

            information_instructmodels = "<p>Instruct models require the maximum output tokens to be specified. This should be roughly half the maximum tokens the model supports.</p>"

            information_noapikey = """
            <p>To use this provider you need <a href="https://platform.openai.com/account/api-keys">an OpenAI API key</a>.</p>
            <p>Api Base can usually be left blank unless you are using a custom OpenAI instance, when you will need to provide the base URL.</p>
            <p>Note that if your API Key is attached to a free trial account the <a href="https://platform.openai.com/docs/guides/rate-limits?context=tier-free">rate limit</a> for requests will be <i>severely</i> limited.</p>
            """

            def __init__(self, settings : SettingsType):
                super().__init__(self.name, SettingsType({
                    "api_key": settings.get_str('api_key', os.getenv('OPENAI_API_KEY')),
                    "api_base": settings.get_str('api_base', os.getenv('OPENAI_API_BASE')),
                    "model": settings.get_str('model', os.getenv('OPENAI_MODEL', "gpt-5-mini")),
                    'temperature': settings.get_float('temperature', env_float('OPENAI_TEMPERATURE', 0.0)),
                    'rate_limit': settings.get_float('rate_limit', env_float('OPENAI_RATE_LIMIT')),
                    "free_plan": settings.get_bool('free_plan', os.getenv('OPENAI_FREE_PLAN') == "True"),
                    'max_instruct_tokens': settings.get_int('max_instruct_tokens', int(os.getenv('MAX_INSTRUCT_TOKENS', '2048'))),
                    'use_httpx': settings.get_bool('use_httpx', os.getenv('OPENAI_USE_HTTPX', "False") == "True"),
                    'reasoning_effort': settings.get_str('reasoning_effort', os.getenv('OPENAI_REASONING_EFFORT', "low")),
                    'stream_responses': settings.get_bool('stream_responses', os.getenv('OPENAI_STREAM_RESPONSES', "True") == "True"),
                }))

                self.refresh_when_changed = ['api_key', 'api_base', 'model']

                self.valid_model_types = [ "gpt", "o1", "o3", "o4" ]
                self.excluded_model_types = [ "vision", "image", "realtime", "audio", "instruct", "search", "transcribe", "tts", "deep-research", "codex" ]
                self.non_reasoning_models = [ "gpt-3", "gpt-4", "gpt-5-chat" ]

            @property
            def api_key(self) -> str|None:
                return self.settings.get_str( 'api_key')

            @property
            def api_base(self) -> str|None:
                return self.settings.get_str( 'api_base')

            @property
            def is_instruct_model(self) -> bool:
                return self.selected_model is not None and self.model_is_instruct_model(self.selected_model)

            @property
            def is_reasoning_model(self) -> bool:
                return self.selected_model is not None and self.model_is_reasoning_model(self.selected_model)

            def model_is_instruct_model(self, model_name : str) -> bool:
                return model_name.find("instruct") >= 0

            def model_is_reasoning_model(self, model_name : str) -> bool:
                return not any(model_name.startswith(model) for model in self.non_reasoning_models)

            def GetTranslationClient(self, settings : SettingsType) -> TranslationClient:
                client_settings = SettingsType(self.settings.copy())
                client_settings.update(settings)
                if self.is_instruct_model:
                    raise ProviderError("Instruct models are no longer supported", provider=self)
                elif self.is_reasoning_model:
                    return OpenAIReasoningClient(client_settings)
                else:
                    return ChatGPTClient(client_settings)

            def GetOptions(self, settings : SettingsType) -> GuiSettingsType:
                options : GuiSettingsType = {
                    'api_key': (str, _("An OpenAI API key is required to use this provider (https://platform.openai.com/account/api-keys)")),
                    'api_base': (str, _("The base URL to use for requests - leave as default unless you know you need something else")),
                }

                if self.api_base:
                    options['use_httpx'] = (bool, _("Use the httpx library for requests. May help if you receive a 307 redirect error with a custom api_base"))

                if self.api_key:
                    models = self.available_models
                    if models:
                        options.update({
                            'model': (models, _("AI model to use as the translator") if models else _("Unable to retrieve models")),
                            'rate_limit': (float, _("Maximum OpenAI API requests per minute. Mainly useful if you are on the restricted free plan"))
                        })

                        model = settings.get_str('model') or self.selected_model or "gpt-5-mini"
                        if self.model_is_instruct_model(model):
                            options['max_instruct_tokens'] = (int, _("Maximum tokens a completion can contain (only applicable for -instruct models)"))

                        if self.model_is_reasoning_model(model):
                            options['reasoning_effort'] = (["minimal", "low", "medium", "high"], _("The level of reasoning effort to use for the model"))
                            options['stream_responses'] = (bool, _("Stream translations in realtime as they are generated"))
                        else:
                            options['temperature'] = (float, _("Amount of random variance to add to translations. Generally speaking, none is best"))

                    else:
                        options['model'] = ([_("Unable to retrieve models")], _("Check API key and base URL and try again"))

                return options

            def GetAvailableModels(self) -> list[str]:
                """
                Returns a list of possible values for the LLM model
                """
                try:
                    if not hasattr(openai, "OpenAI"):
                        raise ProviderError("The OpenAI library is out of date and must be updated", provider=self)

                    if not self.api_key:
                        logging.debug("No OpenAI API key provided")
                        return []

                    client = openai.OpenAI(
                        api_key=self.api_key,
                        base_url=self.api_base or None
                    )
                    response = client.models.list()

                    if not response or not response.data:
                        return []

                    model_list = [ model.id for model in response.data if any(model.id.startswith(prefix) for prefix in self.valid_model_types) ]

                    model_list = [ model for model in model_list if not any([ model.find(exclude) >= 0 for exclude in self.excluded_model_types ]) ]

                    # Maybe this isn't really an OpenAI endpoint, just return all the available models
                    if not model_list:
                        model_list = [ model.id for model in response.data ]

                    return sorted(model_list)

                except Exception as e:
                    logging.error(_("Unable to retrieve available AI models: {error}").format(error=str(e)))
                    return []

            def GetInformation(self) -> str:
                if not self.api_key:
                    return self.information_noapikey
                if self.is_instruct_model:
                    return self.information + self.information_instructmodels
                return self.information

            def ValidateSettings(self) -> bool:
                """
                Validate the settings for the provider
                """
                if not self.api_key:
                    self.validation_message = _("API Key is required")
                    return False

                return True

            def _allow_multithreaded_translation(self) -> bool:
                """
                If user is on the free plan or has set a rate limit it is better not to try parallel requests
                """
                if self.settings.get_bool( 'free_plan'):
                    return False

                if self.settings.get_float( 'rate_limit', 0.0) != 0.0:
                    return False

                return True

    except ImportError:
        from PySubtrans.Helpers.Localization import _
        logging.info(_("Failed to initialise OpenAI SDK. OpenAI provider will not be available"))


import json
import os
import logging
import threading

from PySubtrans.Helpers import GetOutputPath
from PySubtrans.Helpers.Localization import _
from PySubtrans.Helpers.Parse import ParseNames
from PySubtrans.Options import Options, SettingsType
from PySubtrans.Substitutions import Substitutions
from PySubtrans.SettingsType import SettingsType
from PySubtrans.SubtitleEditor import SubtitleEditor
from PySubtrans.SubtitleError import SubtitleError, TranslationAbortedError
from PySubtrans.SubtitleFormatRegistry import SubtitleFormatRegistry
from PySubtrans.Subtitles import Subtitles

from PySubtrans.SubtitleScene import SubtitleScene
from PySubtrans.SubtitleSerialisation import SubtitleDecoder, SubtitleEncoder
from PySubtrans.SubtitleTranslator import SubtitleTranslator
from PySubtrans.TranslationEvents import TranslationEvents

default_encoding = os.getenv('DEFAULT_ENCODING', 'utf-8')

class SubtitleProject:
    """
    Handles loading, saving and creation of project files for LLM-Subtrans
    """
    DEFAULT_PROJECT_SETTINGS : SettingsType = SettingsType({
        'provider': None,
        'model': None,
        'target_language': None,
        'prompt': None,
        'task_type': None,
        'instructions': None,
        'retry_instructions': None,
        'movie_name': None,
        'description': None,
        'names': None,
        'substitutions': None,
        'substitution_mode': None,
        'include_original': None,
        'add_right_to_left_markers': None,
        'instruction_file': None,
        'format': None
    })
   
    def __init__(self, persistent : bool = False):
        """
        A subtitle translation project.

        Can be initialised from a project file or a subtitle file,
        or manually configured by assigning a SubtitleFile and updating settings if necessary.

        :param persistent: if True, the project will be saved to disk and automatically reloaded next time
        """
        self.subtitles : Subtitles = Subtitles(settings=self.DEFAULT_PROJECT_SETTINGS)
        self.events = TranslationEvents()
        self.projectfile : str|None = None
        self.existing_project : bool = False
        self.needs_writing : bool = False
        self.lock = threading.RLock()

        # By default the project is not persistent, i.e. it will not be saved to a file and automatically reloaded next time
        self.use_project_file : bool = persistent

        # By default the translated subtitles will be written to file
        self.write_translation = True

    @property
    def target_language(self) -> str|None:
        return self.subtitles.settings.get_str('target_language') if self.subtitles else None
    
    @property
    def task_type(self) -> str|None:
        return self.subtitles.settings.get_str('task_type') if self.subtitles else None

    @property
    def movie_name(self) -> str|None:
        return self.subtitles.settings.get_str('movie_name') if self.subtitles else None

    @property
    def any_translated(self) -> bool:
        with self.lock:
            return bool(self.subtitles and self.subtitles.any_translated)

    @property
    def all_translated(self) -> bool:
        with self.lock:
            return bool(self.subtitles and self.subtitles.all_translated)

    @target_language.setter
    def target_language(self, value : str|None) -> None:
        self._set_project_setting('target_language', value)

    @task_type.setter
    def task_type(self, value : str|None) -> None:
        self._set_project_setting('task_type', value)

    @movie_name.setter
    def movie_name(self, value : str|None) -> None:
        self._set_project_setting('movie_name', value)

    def InitialiseProject(self, filepath : str, outputpath : str|None = None, reload_subtitles : bool = False):
        """
        Initialize the project by either loading an existing project file or creating a new one.
        Load the subtitles to be translated, either from the project file or the source file.

        :param filepath: the path to the project or a source subtitle file
        :param outputpath: the path to write the translated subtitles to
        :param reload_subtitles: force reloading subtitles from source file
        """
        filepath = os.path.normpath(filepath)
        sourcepath : str = filepath
        self.projectfile = self.GetProjectFilepath(filepath or "subtitles")

        project_file_exists : bool = os.path.exists(self.projectfile)
        project_settings : SettingsType = SettingsType()

        # If initialised with a project file, we are implicitly using a project file
        if filepath == self.projectfile:
            self.use_project_file = True

        read_project : bool = self.use_project_file and project_file_exists
        load_subtitles : bool = reload_subtitles or not read_project

        if not read_project and not load_subtitles:
            raise SubtitleError(_("No project or subtitles to load"))

        if project_file_exists and not read_project:
            logging.warning(_("Project file {} exists but will not be used").format(self.projectfile))

        if read_project:
            logging.info(_("Loading existing project file {}").format(self.projectfile))

            self.ReadProjectFile(self.projectfile)
            project_settings = self.GetProjectSettings()

            subtitles : Subtitles = self.subtitles
            if subtitles:
                self.UpdateOutputPath()

                outputpath = outputpath or GetOutputPath(self.projectfile, self.target_language, subtitles.file_format)
                sourcepath = subtitles.sourcepath if subtitles.sourcepath else sourcepath               
                logging.info(_("Project file loaded"))

                if subtitles.scenes:
                    self.existing_project = True
                    self.needs_writing = False
                    load_subtitles = reload_subtitles
                    if load_subtitles:
                        logging.info(_("Reloading subtitles from the source file"))

                else:
                    logging.error(_("Unable to read project file, starting afresh"))
                    load_subtitles = True

        if load_subtitles:
            try:
                # (re)load the source subtitle file if required
                self.LoadSubtitleFile(sourcepath)

            except Exception as e:
                logging.error(_("Failed to load subtitle file {}: {}").format(filepath, str(e)))
                raise

        subtitles = self.subtitles

        if not subtitles or not subtitles.has_subtitles:
            raise ValueError(_("No subtitles to translate in {}").format(filepath))

        if outputpath:
            subtitles.outputpath = outputpath
            subtitles.file_format = SubtitleFormatRegistry.get_format_from_filename(outputpath)
            self.needs_writing = self.use_project_file

        # Re-apply any project settings and update for compatibility
        if read_project:
            self.UpdateProjectSettings(project_settings)

    def UpdateProjectSettings(self, settings: SettingsType) -> None:
        """
        Update the project settings with validation and filtering
        """
        if isinstance(settings, Options):
            settings = SettingsType(settings)

        with self.lock:
            if not self.subtitles:
                return

            # Update obsolete settings to maintain compatibility
            self._update_compatibility(settings)

            # Filter settings to only include known project settings
            filtered_settings = SettingsType({key: settings[key] for key in settings if key in self.DEFAULT_PROJECT_SETTINGS})

            # Process names and substitutions into standard formats
            if 'names' in filtered_settings:
                names_list = filtered_settings.get('names', [])
                filtered_settings['names'] = ParseNames(names_list)

            if 'substitutions' in filtered_settings:
                substitutions_list = filtered_settings.get('substitutions', [])
                if substitutions_list:
                    filtered_settings['substitutions'] = Substitutions.Parse(substitutions_list)

            # Check if there are any actual changes
            common_keys = filtered_settings.keys() & self.subtitles.settings.keys()
            new_keys = filtered_settings.keys() - self.subtitles.settings.keys()

            if new_keys or not all(filtered_settings.get(key) == self.subtitles.settings.get(key) for key in common_keys):
                self.subtitles.UpdateSettings(filtered_settings)
                self.needs_writing = self.use_project_file and bool(self.subtitles.scenes)

    def UpdateOutputPath(self, path: str|None = None, extension: str|None = None) -> None:
        """
        Set or generate the output path for the translated subtitles
        """
        path = path or self.subtitles.sourcepath
        extension = extension or self.subtitles.file_format
        if not extension:
            extension = SubtitleFormatRegistry.get_format_from_filename(path) if path else None
            extension = extension or '.srt'

        if extension == ".subtrans":
            raise SubtitleError("Cannot use .subtrans as output format")

        outputpath = GetOutputPath(path, self.target_language, extension)
        self.subtitles.outputpath = outputpath
        self.subtitles.file_format = extension

    def SaveOriginal(self, outputpath : str|None = None):
        """
        Write the original subtitles to a file
        """
        try:
            with self.lock:
                outputpath = outputpath or GetOutputPath(self.subtitles.sourcepath, None, self.subtitles.file_format)
                self.subtitles.SaveOriginal(outputpath)

        except Exception as e:
            logging.error(_("Unable to save original subtitles: {}").format(e))

    def SaveTranslation(self, outputpath : str|None = None):
        """
        Write output file
        """
        try:
            with self.lock:
                self.subtitles.SaveTranslation(outputpath)

        except Exception as e:
            logging.error(_("Unable to save translation: {}").format(e))

    def GetEditor(self) -> SubtitleEditor:
        """
        Return a SubtitleEditor that marks the project as needing to be written
        when edits complete successfully.
        """
        if not self.subtitles:
            raise SubtitleError("Cannot edit project without subtitles")

        def mark_project_dirty(success: bool) -> None:
            if success:
                self.needs_writing = True

        return SubtitleEditor(self.subtitles, mark_project_dirty)

    def GetProjectFilepath(self, filepath : str) -> str:
        """
        Calculate the project file path based on the source file path
        """
        path, ext = os.path.splitext(filepath)
        filepath = filepath if ext == '.subtrans' else f"{path}.subtrans"
        return os.path.normpath(filepath)

    def GetBackupFilepath(self, filepath : str) -> str:
        """
        Get the backup file path for the project file
        """
        projectfile = self.GetProjectFilepath(filepath)
        return f"{projectfile}-backup"

    def LoadSubtitleFile(self, filepath: str) -> Subtitles:
        """
        Load subtitles from a file, auto-detecting the format by extension
        """
        with self.lock:
            # Pass default settings for new subtitle files
            self.subtitles = Subtitles(filepath, settings=self.DEFAULT_PROJECT_SETTINGS)
            self.subtitles.LoadSubtitles()

        return self.subtitles

    def SaveProject(self):
        """
        Save the project file or translation file as needed
        """
        with self.lock:
            if self.needs_writing:
                if self.use_project_file:
                    self.UpdateProjectFile()
                if self.any_translated and self.write_translation:
                    self.SaveTranslation()
                self.needs_writing = False

    def UpdateProjectFile(self) -> None:
        """
        Save the project file if it needs updating
        """
        with self.lock:
            if self.needs_writing and self.subtitles and self.subtitles.scenes:
                self.SaveProjectFile()

    def SaveProjectFile(self, projectfile : str|None = None) -> None:
        """
        Write a set of subtitles to a project file
        """
        with self.lock:
            if not self.subtitles:
                raise Exception("Can't write project file, no subtitles")

            if not isinstance(self.subtitles, Subtitles):
                raise Exception("Can't write project file, wrong content type")

            if not self.subtitles.scenes:
                raise Exception("Can't write project file, no scenes")

            if not projectfile:
                projectfile = self.projectfile
            elif projectfile and not self.projectfile:
                self.projectfile = self.GetProjectFilepath(projectfile)

            if not projectfile:
                raise Exception("No file path provided")

            self.WriteProjectToFile(projectfile, encoder_class=SubtitleEncoder)

            self.needs_writing = False

    def SaveBackupFile(self) -> None:
        """
        Save a backup copy of the project
        """
        with self.lock:
            if self.subtitles and self.projectfile:
                backupfile = self.GetBackupFilepath(self.projectfile)
                self.WriteProjectToFile(backupfile, encoder_class=SubtitleEncoder)

    def ReadProjectFile(self, filepath : str|None = None) -> Subtitles|None:
        """
        Load scenes, subtitles and context from a project file
        """
        try:
            filepath = filepath or self.projectfile
            if not filepath:
                raise ValueError(_("No project file path provided"))

            with self.lock:
                logging.info(_("Reading project data from {}").format(str(filepath)))

                with open(filepath, 'r', encoding=default_encoding, newline='') as f:
                    self.subtitles: Subtitles = json.load(f, cls=SubtitleDecoder)

                with SubtitleEditor(self.subtitles) as editor:
                    editor.Sanitise()

                return self.subtitles

        except FileNotFoundError:
            logging.error(_("Project file {} not found").format(filepath))
            return None

        except json.JSONDecodeError as e:
            logging.error(_("Error decoding JSON file: {}").format(e))
            return None

    def GetProjectSettings(self) -> SettingsType:
        """
        Return a dictionary of non-empty settings from the project file
        """
        if not self.subtitles:
            return SettingsType()

        return SettingsType({ key : value for key, value in self.subtitles.settings.items() if value is not None and (value != '' or isinstance(value, list)) })

    def WriteProjectToFile(self, projectfile: str, encoder_class: type|None = None) -> None:
        """
        Save the project settings to a JSON file
        """
        if encoder_class is None:
            raise ValueError("No encoder provided")

        projectfile = os.path.normpath(projectfile)
        logging.info(_("Writing project data to {}").format(str(projectfile)))

        with self.lock:
            with open(projectfile, 'w', encoding=default_encoding) as f:
                project_json = json.dumps(self.subtitles, cls=encoder_class, ensure_ascii=False, indent=4) # type: ignore
                f.write(project_json)

    def TranslateSubtitles(self, translator : SubtitleTranslator) -> None:
        """
        One-stop shop: Use *translator* to translate a project, then save the translation.
        """
        if not self.subtitles:
            raise ValueError("No subtitles to translate")

        if not translator:
            raise ValueError("No translator supplied")

        # Prime new project files
        self.UpdateProjectFile()

        save_translation : bool = self.write_translation and not translator.preview

        translator.events.preprocessed.connect(self._on_preprocessed)
        translator.events.batch_translated.connect(self._on_batch_translated)
        translator.events.scene_translated.connect(self._on_scene_translated)
        translator.events.connect_default_loggers()

        try:
            translator.TranslateSubtitles(self.subtitles)

            if save_translation and not translator.aborted:
                self.SaveTranslation()

        except TranslationAbortedError:
            logging.info(_("Translation aborted"))

        except Exception as e:
            if save_translation and self.subtitles.any_translated:
                logging.warning(_("Translation failed, saving partial results"))
                self.SaveTranslation()

            logging.error(_("Failed to translate subtitles: {}").format(str(e)))
            raise

        finally:
            translator.events.preprocessed.disconnect(self._on_preprocessed)
            translator.events.batch_translated.disconnect(self._on_batch_translated)
            translator.events.scene_translated.disconnect(self._on_scene_translated)
            translator.events.disconnect_default_loggers()

    def TranslateScene(self, translator : SubtitleTranslator, scene_number : int, batch_numbers : list[int]|None = None, line_numbers : list[int]|None = None) -> SubtitleScene|None:
        """
        Pass batches of subtitles to *translator* for translation.
        """
        if not self.subtitles:
            raise ValueError("No subtitles to translate")

        if not translator:
            raise ValueError("No translator supplied")

        translator.events.preprocessed.connect(self._on_preprocessed)
        translator.events.batch_translated.connect(self._on_batch_translated)
        translator.events.connect_default_loggers()

        try:
            scene : SubtitleScene = self.subtitles.GetScene(scene_number)

            scene.errors = []

            translator.TranslateScene(self.subtitles, scene, batch_numbers=batch_numbers, line_numbers=line_numbers)

            return scene

        except TranslationAbortedError:
            pass

        finally:
            translator.events.preprocessed.disconnect(self._on_preprocessed)
            translator.events.batch_translated.disconnect(self._on_batch_translated)
            translator.events.disconnect_default_loggers()

    def _set_project_setting(self, setting_name, value):
        """
        Set a project setting and mark the project as needing to be written if it changes
        """
        with self.lock:
            if not self.subtitles:
                raise SubtitleError(_("Cannot set {setting_name}, no subtitles loaded").format(setting_name=setting_name))

            if self.subtitles.settings.get_str(setting_name) != value:
                self.subtitles.settings[setting_name] = value
                self.needs_writing = self.use_project_file

    def _update_compatibility(self, settings: SettingsType) -> None:
        """
        Update settings for compatibility with older versions
        """
        if not settings.get('description') and settings.get('synopsis'):
            settings['description'] = settings.get('synopsis')

        if settings.get('characters'):
            names = settings.get_str_list('names')
            names.extend(settings.get_str_list('characters'))
            settings['names'] = names
            del settings['characters']

        if settings.get('gpt_prompt'):
            settings['prompt'] = settings['gpt_prompt']
            del settings['gpt_prompt']

        if settings.get('gpt_model'):
            settings['model'] = settings['gpt_model']
            del settings['gpt_model']

        if not settings.get('substitution_mode'):
            settings['substitution_mode'] = "Partial Words" if settings.get('match_partial_words') else "Auto"


    def _on_preprocessed(self, sender, scenes) -> None:
        logging.debug("Pre-processing finished")
        self.events.preprocessed.send(self, scenes=scenes)

    def _on_batch_translated(self, sender, batch) -> None:
        logging.debug("Batch translated")
        self.needs_writing = self.use_project_file
        self.events.batch_translated.send(self, batch=batch)

    def _on_scene_translated(self, sender, scene) -> None:
        logging.debug("Scene translated")
        self.needs_writing = self.use_project_file
        self.events.scene_translated.send(self, scene=scene)





from collections.abc import Callable
import functools
import logging
import os
import sys
import unittest
from datetime import datetime
from typing import Any

from PySubtrans.SettingsType import SettingsType
from PySubtrans.SubtitleFormatRegistry import SubtitleFormatRegistry
from PySubtrans.Subtitles import Subtitles

separator = "".center(60, "-")
wide_separator = "".center(120, "-")

def log_info(text: str, prefix: str = ""):
    """
    Logs a string as individual lines with an optional prefix on each line using logging.info.
    """
    for line in text.strip().split("\n"):
        logging.info(f"{prefix}{line}")

def log_error(text: str, prefix: str = ""):
    """
    Logs a string as individual lines with an optional prefix on each line using logging.error.
    """
    for line in text.strip().split("\n"):
        logging.error(f"{prefix}{line}")

def log_test_name(test_name: str):
    """
    Logs the name of the test with a separator before and after.
    """
    logging.info(separator)
    log_info(test_name.center(len(separator)))
    logging.info(separator)

def log_expected_result(expected : Any, result : Any):
    """
    Logs the expected result and the actual result.
    """
    log_info(str(expected), prefix="===".ljust(10))
    log_info(str(result), prefix="-->".ljust(10))
    if expected != result:
        log_error("*** UNEXPECTED RESULT! ***", prefix="!!!".ljust(10))
    logging.info(separator)

def log_input_expected_result(input : Any, expected : Any, result : Any):
    """
    Logs the input value, the expected result and the actual result.
    """
    log_info(str(input), prefix="".ljust(10))
    log_info(str(expected), prefix="===".ljust(10))
    log_info(str(result), prefix="-->".ljust(10))
    if expected != result:
        log_error("*** UNEXPECTED RESULT! ***", prefix="!!!".ljust(10))
    logging.info(separator)

def log_input_expected_error(input : Any, expected_error : type[Exception], result : Any):
    """
    Logs the input value, the expected error and the actual error.
    """
    log_info(str(input), prefix="".ljust(10))
    log_info(expected_error.__name__, prefix="===".ljust(10))
    if not isinstance(result, expected_error):
        log_error("*** UNEXPECTED ERROR! ***", prefix="!!!".ljust(10))
    log_info(str(result), prefix="-->".ljust(10))
    logging.info(separator)


def skip_if_debugger_attached(test_method):
    """
    Decorator to skips a test method when debugger is attached.
    Use this to skip tests that raise expected exceptions when debugging,
    to facilitate debugging tests that are UNEXPECTEDLY raising exceptions.

    Usage:
        @skip_if_debugger_attached
        def test_some_exception_handling(self):
            # Test that raises exceptions
            pass
    """
    @functools.wraps(test_method)
    def wrapper(self, *args, **kwargs):
        if sys.gettrace() is not None:
            test_name = test_method.__name__
            logging.info(f"Skipped {test_name} when debugger is attached")
        else:
            return test_method(self, *args, **kwargs)
    return wrapper

def create_logfile(results_dir : str, log_name : str, log_level = logging.DEBUG) -> logging.FileHandler:
    """
    Creates a log file with the specified name in the specified directory and adds it to the root logger.
    """
    log_path = os.path.join(results_dir, log_name)
    file_handler = logging.FileHandler(log_path, encoding='utf-8', mode='w')
    file_handler.setLevel(log_level)
    file_handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logging.getLogger('').addHandler(file_handler)
    return file_handler

def end_logfile(file_handler : logging.FileHandler):
    """
    Closes the file handler for the log file.
    """
    logging.getLogger('').removeHandler(file_handler)
    file_handler.close()

def _configure_base_logger(results_path, test_name):
    """
    Configures and returns a base logger that logs DEBUG messages to the console
    and to a general log file named after the test name.
    """
    logger = logging.getLogger(test_name)
    logger.setLevel(logging.DEBUG)

    test_log_path = os.path.join(results_path, f"{test_name}.log")
    file_handler = logging.FileHandler(test_log_path, mode='w', encoding='utf-8')
    file_formatter = logging.Formatter('%(message)s')
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    return logger

def _add_test_file_logger(logger, results_path, input_filename, test_name):
    """
    Adds a file handler to log INFO level messages to a specific file named after the input file (without extension) and test name.
    """
    base_filename, dummy = os.path.splitext(input_filename) # type: ignore[ignore-unused]
    input_log_path = os.path.join(results_path, f"{base_filename}-{test_name}.log")
    file_handler = logging.FileHandler(input_log_path, mode='w', encoding='utf-8')
    file_formatter = logging.Formatter('%(message)s')
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.INFO)
    logger.addHandler(file_handler)
    return file_handler

def RunTestOnAllSubtitleFiles(run_test : Callable, test_options: list[dict], directory_path: str, results_path: str|None = None):
    """
    Run a series of tests on all .srt files in the test_subtitles directory.
    """
    test_name = run_test.__name__

    results_path = results_path or directory_path
    os.makedirs(results_path, exist_ok=True)

    logger = _configure_base_logger(results_path, test_name)

    print(separator)
    print(f"Running {test_name}")

    logger.info(separator)
    logger.info(f"Running {test_name}")
    logger.info(separator)
    logger.info("")

    supported_formats = SubtitleFormatRegistry.enumerate_formats()

    def _is_supported_subtitle_file(f):
        file_path = os.path.join(directory_path, f)
        return (
            os.path.isfile(file_path)
            and SubtitleFormatRegistry.get_format_from_filename(f) in supported_formats
        )

    files = [f for f in os.listdir(directory_path) if _is_supported_subtitle_file(f)]
    print (f"Running {test_name} on {len(files)} files in {directory_path}...")

    for file in files:
        file_handler = _add_test_file_logger(logger, results_path, file, test_name)

        filepath = os.path.join(directory_path, file)

        current_time = datetime.now().strftime("%Y-%m-%d at %H:%M")
        logger.info(f"File: {filepath}")
        logger.info(f"Tested: {current_time}")
        logger.info(separator)

        try:
            subtitles = Subtitles(filepath)
            subtitles.LoadSubtitles()

            for options in test_options:
                logger.info("")
                run_test(subtitles, logger, SettingsType(options))

        except Exception as e:
            logger.error(f"Error processing {filepath}: {str(e)}")
            print(f"!!! ERROR RUNNING {test_name} ON {file} !!!")

        finally:
            logger.removeHandler(file_handler)





import logging
from typing import cast
from PySubtrans.Options import Options, SettingsType
from PySubtrans.SettingsType import GuiSettingsType, SettingsType
from PySubtrans.TranslationClient import TranslationClient

class TranslationProvider:
    """
    Base class for translation service providers.
    """
    def __init__(self, name : str, settings : SettingsType):
        self.name : str = name
        self.settings : SettingsType = settings
        self._available_models : list[str] = []
        self.refresh_when_changed : list[str] = []
        self.validation_message : str|None = None

    @property
    def available_models(self) -> list[str]:
        """
        list of available models for the provider
        """
        if not self._available_models:
            self._available_models = self.GetAvailableModels()

        return self._available_models

    @property
    def all_available_models(self) -> list[str]:
        """
        Returns all available models for the provider, including those currently filtered out
        """
        return self.available_models

    @property
    def selected_model(self) -> str|None:
        """
        The currently selected model for the provider
        """
        name : str|None = self.settings.get_str( 'model')
        return name.strip() if name else None

    @property
    def allow_multithreaded_translation(self) -> bool:
        """
        Returns True if the provider supports multithreaded translation
        """
        return self._allow_multithreaded_translation()

    def GetAvailableModels(self) -> list[str]:
        """
        Returns a list of possible model for the provider
        """
        raise NotImplementedError

    def ResetAvailableModels(self):
        """
        Reset the available models for the provider
        """
        self._available_models = []

    def GetInformation(self) -> str|None:
        """
        Returns information about the provider settings
        """
        return None

    def GetTranslationClient(self, settings : SettingsType) -> TranslationClient:
        """
        Returns a new instance of the translation client for this provider
        """
        raise NotImplementedError

    def GetOptions(self, settings : SettingsType) -> GuiSettingsType:
        """
        Returns the configurable options for the provider
        """
        raise NotImplementedError

    def ValidateSettings(self) -> bool:
        """
        Validate the settings for the provider
        """
        return True

    def UpdateSettings(self, settings : SettingsType):
        """
        Update the settings for the provider
        """
        if isinstance(settings, Options):
            options = cast(Options, settings)
            options.InitialiseProviderSettings(self.name, self.settings)
            settings = options.provider_settings[self.name]

        # Update the settings
        for k, v in settings.items():
            if k in self.settings:
                self.settings[k] = v

    def GetCombinedSettings(self, overrides : SettingsType) -> SettingsType:
        """
        Merge saved settings with default settings to ensure all keys are present
        """
        combined_settings = SettingsType(self.settings.copy())
        combined_settings.update(overrides)
        return combined_settings

    def _allow_multithreaded_translation(self) -> bool:
        """
        Returns True if the provider supports multithreaded translation
        """
        return False

    @classmethod
    def get_providers(cls) -> dict:
        """
        Return a dictionary of all available providers
        """
        if not cls.__subclasses__():
            # Import the providers package, which will trigger explicit imports
            from . import Providers  # type: ignore[ignore-unused]

        providers = { cast(TranslationProvider, provider).name : provider for provider in cls.__subclasses__() }

        return providers

    @classmethod
    def get_provider(cls, options : Options):
        """
        Create a new instance of the provider with the given name
        """
        if not isinstance(options, Options):
            raise ValueError("Options object required")

        if not options.provider:
            raise ValueError("No provider set")

        provider_settings = options.current_provider_settings or SettingsType()

        translation_provider : TranslationProvider = cls.create_provider(options.provider, provider_settings)
        if not translation_provider:
            raise ValueError(f"Unable to create translation provider '{options.provider}'")

        translation_provider.UpdateSettings(options)

        return translation_provider

    @classmethod
    def create_provider(cls, name, provider_settings):
        providers = cls.get_providers().items()
        for provider_name, provider in providers:
            if provider_name == name:
                return provider(provider_settings)

        raise ValueError(f"Unknown translation provider: {name}")


    @classmethod
    def get_available_models(cls, options : Options):
        """ Get the available models for the selected provider """
        if not isinstance(options, Options):
            raise ValueError("Options object required")

        if not options.provider:
            return []

        provider_class = cls.get_provider(options)
        if not provider_class:
            return []

        return provider_class.available_models


import importlib.util
import logging

from PySubtrans.Helpers.Localization import _
from PySubtrans.Options import env_float, env_int, env_bool
from PySubtrans.SettingsType import GuiSettingsType, SettingsType

if not importlib.util.find_spec("anthropic"):
    logging.debug(_("Anthropic SDK is not installed. Claude provider will not be available"))
else:
    try:
        import anthropic
        import os

        from copy import deepcopy

        from PySubtrans.Helpers.Localization import _
        from PySubtrans.Providers.Clients.AnthropicClient import AnthropicClient
        from PySubtrans.TranslationClient import TranslationClient
        from PySubtrans.TranslationProvider import TranslationProvider
        from PySubtrans.Options import SettingsType

        class Provider_Claude(TranslationProvider):
            name = "Claude"

            information = """
            <p>Select the <a href="https://docs.anthropic.com/claude/docs/models-overview">AI model</a> to use as a translator.</p>
            <p>Note that each model has a <a href="https://docs.anthropic.com/claude/docs/models-overview">maximum tokens limit</a>.</p>
            <p>See the <a href="https://docs.anthropic.com/claude/reference/rate-limits">Anthropic documentation</a> for information on rate limits and costs</p>
            """

            information_noapikey = """
            <p>To use Claude you need to provide an <a href="https://console.anthropic.com/settings/keys">Anthropic API Key </a>.</p>
            """

            default_model = "claude-3-5-haiku-latest"

            def __init__(self, settings : SettingsType):
                super().__init__(self.name, SettingsType({
                    "api_key": settings.get_str('api_key') or os.getenv('CLAUDE_API_KEY'),
                    "model": settings.get_str('model') or os.getenv('CLAUDE_MODEL', self.default_model),
                    'stream_responses': settings.get_bool('stream_responses', os.getenv('CLAUDE_STREAM_RESPONSES', "True") == "True"),
                    "thinking": settings.get_bool('thinking', False),
                    "max_tokens": settings.get_int('max_tokens') or env_int('CLAUDE_MAX_TOKENS', 4096),
                    "max_thinking_tokens": settings.get_int('max_thinking_tokens') or env_int('CLAUDE_MAX_THINKING_TOKENS', 1024),
                    'temperature': settings.get_float('temperature', env_float('CLAUDE_TEMPERATURE', 0.0)),
                    'rate_limit': settings.get_float('rate_limit', env_float('CLAUDE_RATE_LIMIT', 10.0)),
                    'proxy': settings.get_str('proxy') or os.getenv('CLAUDE_PROXY'),
                }))

                self.refresh_when_changed = ['api_key', 'model', 'thinking']

                self.claude_models = []

            @property
            def api_key(self) -> str|None:
                return self.settings.get_str( 'api_key')
            
            @property
            def allow_thinking(self) -> bool:
                return self.settings.get_bool( 'thinking', False)
            
            @property
            def max_tokens(self) -> int:
                return self.settings.get_int( 'max_tokens') or 8192
            
            @property
            def max_thinking_tokens(self) -> int:
                return self.settings.get_int( 'max_thinking_tokens') or 1024

            def GetTranslationClient(self, settings : SettingsType) -> TranslationClient:
                client_settings : dict = deepcopy(self.settings)
                client_settings.update(settings)
                client_settings.update({
                    'model': self._get_model_id(self.selected_model) if self.selected_model else None,
                    'supports_streaming': True,
                    'supports_conversation': True,
                    'supports_system_messages': False,
                    'supports_system_prompt': True
                    })
                return AnthropicClient(client_settings)

            def GetAvailableModels(self) -> list[str]:
                if not self.api_key:
                    return []
                
                if not self.claude_models:
                    self.claude_models = self._get_claude_models()

                models = [model.display_name for model in self.claude_models]

                return models

            def RefreshAvailableModels(self):
                self._available_models = self.GetAvailableModels()

            def GetInformation(self):
                return self.information if self.api_key else self.information_noapikey

            def GetOptions(self, settings : SettingsType) -> GuiSettingsType:
                options : GuiSettingsType = {
                    'api_key': (str, _("An Anthropic Claude API key is required to use this provider (https://console.anthropic.com/settings/keys)"))
                    }

                if not self.api_key:
                    return options

                self.RefreshAvailableModels()

                if self.available_models:
                    options.update({
                        'model': (self.available_models, _("The model to use for translations")),
                        'stream_responses': (bool, _("Stream translations in realtime as they are generated")),
                        'temperature': (float, _("The temperature to use for translations (default 0.0)")),
                        'rate_limit': (float, _("The rate limit to use for translations (default 60.0)")),
                        'max_tokens': (int, _("The maximum number of tokens to use for translations")),
                        'thinking': (bool, _("Enable thinking mode for translations")),
                    })

                if self.allow_thinking:
                    options['max_thinking_tokens'] = (int, _("The maximum number of tokens to use for thinking"))

                options['proxy'] = (str, _("Optional proxy server to use for requests (e.g. https://api.not-anthropic.com/"))
                return options

            def _allow_multithreaded_translation(self) -> bool:
                """
                If user has set a rate limit don't attempt parallel requests to make sure we respect it
                """
                if self.settings.get_float( 'rate_limit', 0.0) != 0.0:
                    return False

                return True

            def _get_claude_models(self):
                if not self.api_key:
                    return []

                try:
                    client = anthropic.Anthropic(api_key=self.api_key)
                    model_list = client.models.list()

                    return [ m for m in model_list if m.type == 'model' ]

                except Exception as e:
                    logging.error(_("Unable to retrieve Claude model list: {error}").format(
                        error=str(e)
                    ))
                    return []

            def _get_model_id(self, name : str) -> str:
                if not self.claude_models:
                    self.claude_models = self._get_claude_models()

                for m in self.claude_models:
                    if m.id == name or m.display_name == name:
                        return m.id

                raise ValueError(f"Model {name} not found")

    except ImportError:
        logging.info(_("Unable to initialise Anthropic SDK. Claude provider will not be available"))



import os
import datetime
import logging
import requests

from PySubtrans.version import __version__
from PySubtrans.Helpers.Resources import config_dir

repo_name = "llm-subtrans"
repo_owner = "machinewrapped"

last_check_file = os.path.join(config_dir, 'last_check.txt')

def CheckIfUpdateAvailable():
    try:
        url = f"https://api.github.com/repos/{repo_owner}/{repo_name}/releases/latest"
        response = requests.get(url)

        if response.status_code == 200:
            with open(last_check_file, "w") as f:
                f.write(datetime.date.today().isoformat())

            latest_version = response.json()["tag_name"]

            if latest_version != __version__:
                logging.info(f"A new version ({latest_version}) of {repo_name} is available!")
                return True

            else:
                logging.debug(f"You have the latest version ({__version__}) of {repo_name}.")
        else:
            logging.debug(f"Failed to get latest release of {repo_name}. Error: {response.status_code}")

    except Exception as e:
        logging.debug(f"Unable to check if an update is available: {str(e)}")

    return False

def CheckIfUpdateCheckIsRequired():
    if not os.path.exists(last_check_file):
        return True

    try:
        with open(last_check_file, "r") as f:
            last_check = datetime.date.fromisoformat(f.read().strip())

        return datetime.date.today() > last_check

    except FileNotFoundError:
        return True




import os
from copy import deepcopy

from PySubtrans.Helpers.Localization import _
from PySubtrans.Options import MULTILINE_OPTION, SettingsType, env_bool, env_float, env_int
from PySubtrans.Providers.Clients.CustomClient import CustomClient
from PySubtrans.SettingsType import GuiSettingsType, SettingsType
from PySubtrans.TranslationClient import TranslationClient
from PySubtrans.TranslationPrompt import default_prompt_template
from PySubtrans.TranslationProvider import TranslationProvider

class Provider_CustomServer(TranslationProvider):
    name = "Custom Server"

    information = """
    <p>This provider allows you to connect to a server running locally (or otherwise accessible) with an OpenAI compatible API,
    e.g. <a href="https://lmstudio.ai/">LM Studio</a>.</p>
    <p>The AI model to use as a translator should be specified on the server, which should then provide an address and supported endpoints.</p>
    <p>Completion and chat endpoints are supported. For chat endpoints, indicate whether the server supports system messages.</p>
    <p>Note that locally hosted models are likely to be much less capable than the large models hosted in the cloud, and results may be underwhelming.</p>
    """

    information_invalid = """
    <p>This provider allows you to connect to a server running locally (or otherwise accessible) with an OpenAI compatible API,
    e.g. <a href="https://lmstudio.ai/">LM Studio</a>.</p>
    <p><b>Server address and endpoint must be provided.</b></p>
    """

    def __init__(self, settings : SettingsType):
        super().__init__(self.name, SettingsType({
            'server_address': settings.get_str('server_address', os.getenv('CUSTOM_SERVER_ADDRESS', "http://localhost:1234")),
            'endpoint': settings.get_str('endpoint', os.getenv('CUSTOM_ENDPOINT', "/v1/chat/completions")),
            'supports_conversation': settings.get_bool('supports_conversation', env_bool('CUSTOM_SUPPORTS_CONVERSATION', True)),
            'supports_system_messages': settings.get_bool('supports_system_messages', env_bool('CUSTOM_SUPPORTS_SYSTEM_MESSAGES', True)),
            'prompt_template': settings.get_str('prompt_template', os.getenv('CUSTOM_PROMPT_TEMPLATE', default_prompt_template)),
            'temperature': settings.get_float('temperature', env_float('CUSTOM_TEMPERATURE', 0.0)),
            'max_tokens': settings.get_int('max_tokens', env_int('CUSTOM_MAX_TOKENS', 0)),
            'max_completion_tokens': settings.get_int('max_completion_tokens', env_int('CUSTOM_MAX_COMPLETION_TOKENS', 0)),
            'timeout': settings.get_int('timeout', env_int('CUSTOM_TIMEOUT', 300)),
            "api_key": settings.get_str('api_key', os.getenv('CUSTOM_API_KEY')),
            "model": settings.get_str('model', os.getenv('CUSTOM_MODEL')),
            'supports_parallel_threads': settings.get_bool('supports_parallel_threads', env_bool('CUSTOM_SUPPORTS_PARALLEL_THREADS', False))
            }))

        #TODO: Add additional parameters option
        #TODO: Add support for custom response parser
        self.refresh_when_changed = ['server_address', 'supports_conversation']

    @property
    def server_address(self) -> str|None:
        return self.settings.get_str( 'server_address')

    @property
    def endpoint(self) -> str|None:
        return self.settings.get_str( 'endpoint')

    @property
    def api_key(self) -> str|None:
        return self.settings.get_str( 'api_key')

    @property
    def supports_conversation(self) -> bool:
        return self.settings.get_bool( 'supports_conversation', False)

    @property
    def supports_system_messages(self) -> bool:
        return self.settings.get_bool( 'supports_system_messages', False)

    @property
    def prompt_template(self) -> str|None:
        return self.settings.get_str( 'prompt_template')

    def GetTranslationClient(self, settings : SettingsType) -> TranslationClient:
        client_settings : dict = deepcopy(self.settings)
        client_settings.update(settings)
        return CustomClient(client_settings)

    def GetAvailableModels(self) -> list[str]:
        # Choose the model on the server
        return [self.selected_model] if self.selected_model else []

    def GetInformation(self):
        if self.ValidateSettings():
            return self.information
        else:
            return self.information_invalid

    def GetOptions(self, settings : SettingsType) -> GuiSettingsType:
        options : GuiSettingsType = {
            'server_address': (str, _("The address of the server")),
            'endpoint': (str, _("The API function to call on the server")),
        }

        if self.ValidateSettings():
            options['supports_conversation'] = (bool, _("Attempt to communicate with the endpoint using chat format"))

            if self.settings.get_bool( 'supports_conversation'):
                options['supports_system_messages'] = (bool, _("Instructions will be sent as system messages rather than the user prompt"))

            options.update({
                'prompt_template': (MULTILINE_OPTION, _("Template for the prompt to send to the server (use {prompt} and {context} tags)")),
                'temperature': (float, _("Higher temperature introduces more randomness to the translation (default 0.0)")),
                'max_tokens': (int, _("The maximum number of tokens the AI should generate in the response (0 for unlimited)")),
                'max_completion_tokens': (int, _("Alternative to max_tokens for some servers")),
                'timeout': (int, _("Timeout for the request in seconds (default 300)")),
                'api_key': (str, _("API key if needed (this is normally not needed for a local server)")),
                'model': (str, _("The model ID (for local servers this is usually not required")),
                'supports_parallel_threads': (bool, _("Use parallel threads for translation requests (may be faster but may not work with the server)"))
            })

        return options

    def ValidateSettings(self) -> bool:
        """
        Validate the settings for the provider
        """
        if not self.server_address:
            self.validation_message = "Server address must be provided"
            return False

        if not self.endpoint:
            self.validation_message = "Endpoint must be provided"
            return False

        return True

    def _allow_multithreaded_translation(self) -> bool:
        """
        User can decide whether to use parallel threads with their model
        """
        return self.settings.get_bool( 'supports_parallel_threads', False)



from __future__ import annotations

from typing import Any

from PySubtrans.SubtitleLine import SubtitleLine


class SubtitleData:
    """
    Format-agnostic container for subtitle lines and file-level metadata.

    Attributes:
        lines (list[SubtitleLine]): List of subtitle lines in the LLM-Subtrans internal representation
        metadata (dict[str, Any]): File-level metadata extracted from or required by specific formats
        start_line_number (int|None): Optional base line number for formats that support line numbering
        detected_format (str|None): Optional detected file format/extension (e.g. '.srt')
    """

    def __init__(self, lines : list[SubtitleLine]|None = None, metadata : dict[str, Any]|None = None, start_line_number : int|None = None, detected_format : str|None = None
    ):
        self.lines : list[SubtitleLine] = lines or []
        self.metadata : dict[str, Any] = metadata or {}
        self.start_line_number : int|None = start_line_number
        self.detected_format : str|None = detected_format




import logging

import boto3 # type: ignore[import]

from PySubtrans.Helpers import FormatMessages
from PySubtrans.Helpers.Localization import _
from PySubtrans.Options import SettingsType

from PySubtrans.Translation import Translation
from PySubtrans.TranslationClient import TranslationClient
from PySubtrans.TranslationPrompt import TranslationPrompt
from PySubtrans.TranslationRequest import TranslationRequest
from PySubtrans.SubtitleError import TranslationImpossibleError, TranslationResponseError

class BedrockClient(TranslationClient):
    """
    Handles communication with Amazon Bedrock to request translations
    """
    def __init__(self, settings : SettingsType):
        super().__init__(settings)

        logging.info(_("Translating with Bedrock model {model_id}, using region: {aws_region}").format(
            model_id=self.model_id, aws_region=self.aws_region
        ))

        self.client = boto3.client(
            'bedrock-runtime',
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_access_key,
            region_name=self.aws_region
        )

    @property
    def access_key(self) -> str|None:
        return self.settings.get_str( 'access_key')

    @property
    def secret_access_key(self) -> str|None:
        return self.settings.get_str( 'secret_access_key')

    @property
    def aws_region(self) -> str|None:
        return self.settings.get_str( 'aws_region')

    @property
    def model_id(self) -> str|None:
        return self.settings.get_str( 'model')

    @property
    def max_tokens(self) -> int:
        return self.settings.get_int( 'max_tokens') or 4096

    def _request_translation(self, request: TranslationRequest, temperature: float|None = None) -> Translation|None:
        """
        Request a translation based on the provided prompt
        """
        if not self.access_key:
            raise TranslationImpossibleError(_("Access key must be set in .env or provided as an argument"))

        if not self.secret_access_key:
            raise TranslationImpossibleError(_("Secret access key must be set in .env or provided as an argument"))

        if not self.aws_region:
            raise TranslationImpossibleError(_("AWS region must be set in .env or provided as an argument"))

        if not self.model_id:
            raise TranslationImpossibleError(_("Model ID must be provided as an argument"))

        if not request.prompt.system_prompt:
            raise TranslationImpossibleError(_("No system prompt provided"))

        logging.debug(f"Messages:\n{FormatMessages(request.prompt.messages)}")

        content = _structure_messages(request.prompt.messages)

        if not content or not isinstance(request.prompt.content, list):
            raise TranslationImpossibleError(_("No content provided for translation"))

        reponse = self._send_messages(request.prompt.system_prompt, content, temperature=temperature)

        translation = Translation(reponse) if reponse else None

        return translation

    def _send_messages(self, system_prompt : str, messages : list[dict], temperature : float|None = None) -> dict|None:
        """
        Make a request to the Amazon Bedrock API to provide a translation
        """
        if self.aborted:
            return None

        try:
            inference_config = {
                    'temperature' : temperature or 0.0,
                    'maxTokens' : self.max_tokens
                }

            if self.supports_system_prompt and system_prompt:
                result = self.client.converse(
                    modelId=self.model_id,
                    messages=messages,
                    system = [{ 'text' : system_prompt }],
                    inferenceConfig = inference_config
                    )
            else:
                result = self.client.converse(
                    modelId=self.model_id,
                    messages=messages,
                    inferenceConfig = inference_config
                    )

            if self.aborted:
                return None

            output = result.get('output')

            if not output:
                raise TranslationResponseError(_("No output returned in the response"), response=result)

            response = {}

            if 'stopReason' in result:
                response['finish_reason'] = result['stopReason']

            if 'usage' in result:
                response['prompt_tokens'] = result['usage'].get('inputTokens')
                response['output_tokens'] = result['usage'].get('outputTokens')
                response['total_tokens'] = result['usage'].get('totalTokens')

            message = output.get('message')
            if message and message.get('role') == 'assistant':
                text = [ content.get('text') for content in message.get('content',[]) ]
                response['text'] = '\n'.join(text)

            # Return the response if the API call succeeds
            return response

        except Exception as e:
            raise TranslationImpossibleError(_("Error communicating with Bedrock: {error}").format(
                error=str(e)
            ), error=e)

def _structure_messages(messages : list[dict[str,str]]) -> list[dict]:
    """
    Structure the messages to be sent to the API
    """
    return [
        {
            'role' : message['role'],
            'content' : [{ 'text': message['content'] }]
        }
        for message in messages]



import json
import logging
import os
import httpx

from PySubtrans.Helpers.Localization import _
from PySubtrans.Options import SettingsType, env_float, env_int
from PySubtrans.Providers.Clients.DeepSeekClient import DeepSeekClient
from PySubtrans.SettingsType import GuiSettingsType, SettingsType
from PySubtrans.TranslationClient import TranslationClient
from PySubtrans.TranslationProvider import TranslationProvider

class DeepSeekProvider(TranslationProvider):
    name = "DeepSeek"

    information = """
    <p>Select the <a href="https://api-docs.deepseek.com/quick_start/pricing">model</a> to use as a translator.</p>
    <p>Not that reasoning models are not generally recommended as translators.</p>
    """

    information_noapikey = """
    <p>To use this provider you need <a href="https://platform.deepseek.com/api_keys">a DeepSeek API key</a>.</p>
    <p>Note that you must have credit to use DeepSeek, there is no free usage tier.</p>
    """

    def __init__(self, settings : SettingsType):
        super().__init__(self.name, SettingsType({
            "api_key": settings.get_str('api_key', os.getenv('DEEPSEEK_API_KEY')),
            "api_base": settings.get_str('api_base', os.getenv('DEEPSEEK_API_BASE', "https://api.deepseek.com")),
            "model": settings.get_str('model', os.getenv('DEEPSEEK_MODEL', "deepseek-chat")),
            'max_tokens': settings.get_int('max_tokens', env_int('DEEPSEEK_MAX_TOKENS', 8192)),
            'temperature': settings.get_float('temperature', env_float('DEEPSEEK_TEMPERATURE', 1.3)),
            'rate_limit': settings.get_float('rate_limit', env_float('DEEPSEEK_RATE_LIMIT')),
            'reuse_client': settings.get_bool('reuse_client', False),
            'endpoint': settings.get_str('endpoint', '/v1/chat/completions'),
            'stream_responses': settings.get_bool('stream_responses', True),
        }))
        self.refresh_when_changed = ['api_key', 'api_base', 'model', 'endpoint']

    @property
    def api_key(self) -> str|None:
        return self.settings.get_str( 'api_key')

    @property
    def api_base(self) -> str|None:
        return self.settings.get_str( 'api_base')
    
    @property
    def server_address(self) -> str|None:
        return self.api_base

    def GetTranslationClient(self, settings : SettingsType) -> TranslationClient:
        client_settings = SettingsType(self.settings.copy())
        client_settings.update(settings)
        return DeepSeekClient(client_settings)

    def GetOptions(self, settings : SettingsType) -> GuiSettingsType:
        options : GuiSettingsType = {
            'api_key': (str, _("A DeepSeek API key is required to use this provider (https://platform.deepseek.com/api_keys)")),
            'api_base': (str, _("The base URL to use for requests (default is https://api.deepseek.com)")),
        }

        if self.api_key:
            models = self.available_models
            if models:
                options.update({
                    'model': (models, _("AI model to use as the translator")),
                    'stream_responses': (bool, _("Enable streaming responses for real-time translation updates")),
                    'reuse_client': (bool, _("Reuse connection for multiple requests (otherwise a new connection is established for each)")),
                    'max_tokens': (int, _("Maximum number of output tokens to return in the response.")),
                    'temperature': (float, _("Amount of random variance to add to translations. Generally speaking, none is best")),
                    'rate_limit': (float, _("Maximum API requests per minute.")),
                })
            else:
                options['model'] = (["Unable to retrieve models"], _("Check API key and base URL and try again"))

        return options

    def GetAvailableModels(self) -> list[str]:
        """
        Fetch available models from DeepSeek API
        """
        if not self.api_key:
            logging.debug("No DeepSeek API key provided")
            return []

        try:
            if not self.server_address:
                logging.debug("No DeepSeek API base URL provided")
                return []

            url = self.server_address.rstrip('/') + '/v1/models'
            headers = {'Authorization': f"Bearer {self.api_key}"} if self.api_key else {}

            with httpx.Client(timeout=15) as client:
                result = client.get(url, headers=headers)
                if result.is_error:
                    logging.error(_("Error fetching models: {status} {text}").format(
                        status=result.status_code, text=result.text))
                    return []

                try:
                    data = result.json()
                    models = [m['id'] for m in data.get('data', [])]
                    return sorted(models)
                except json.JSONDecodeError:
                    logging.error(_("Unable to parse server response as JSON: {response_text}").format(response_text=result.text))
                    return []

        except Exception as e:
            logging.error(_("Unable to retrieve available models: {error}").format(error=str(e)))
            return []

    def GetInformation(self) -> str:
        if not self.api_key:
            return self.information_noapikey
        return self.information

    def ValidateSettings(self) -> bool:
        """
        Validate the settings for the provider
        """
        if not self.api_key:
            self.validation_message = _("API Key is required")
            return False

        return True

    def _allow_multithreaded_translation(self) -> bool:
        """
        If user has set a rate limit we can't make multiple requests at once
        """
        if self.settings.get_float( 'rate_limit', 0.0) != 0.0:
            return False

        return True




import logging
import srt # type: ignore
from collections.abc import Iterator
from typing import TextIO

from PySubtrans.SubtitleFileHandler import (
    SubtitleFileHandler,
    default_encoding,
    fallback_encoding,
)
from PySubtrans.SubtitleLine import SubtitleLine
from PySubtrans.SubtitleData import SubtitleData
from PySubtrans.SubtitleError import SubtitleParseError
from PySubtrans.Helpers.Localization import _

class SrtFileHandler(SubtitleFileHandler):
    """
    File handler for SRT subtitle format.
    Encapsulates all SRT library usage for file I/O operations.
    SRT is a simple format with minimal metadata.
    """
    
    SUPPORTED_EXTENSIONS = {'.srt': 10}

    def load_file(self, path: str) -> SubtitleData:
        try:
            with open(path, 'r', encoding=default_encoding, newline='') as f:
                return self.parse_file(f)
        except UnicodeDecodeError:
            with open(path, 'r', encoding=fallback_encoding, newline='') as f:
                return self.parse_file(f)
    
    def parse_file(self, file_obj: TextIO) -> SubtitleData:
        """
        Parse SRT file content and return SubtitleData with lines and metadata.
        """
        lines = list(self._parse_srt_items(file_obj))
        return SubtitleData(lines=lines, metadata={}, detected_format='.srt')
    
    def parse_string(self, content: str) -> SubtitleData:
        """
        Parse SRT string content and return SubtitleData with lines and metadata.
        """
        lines = list(self._parse_srt_items(content))
        return SubtitleData(lines=lines, metadata={}, detected_format='.srt')

    def compose(self, data: SubtitleData) -> str:
        """
        Compose subtitle lines into SRT format string.
        
        Args:
            data: SubtitleData containing lines and file metadata
            
        Returns:
            str: SRT formatted subtitle content
        """
        from PySubtrans.Helpers.Text import IsRightToLeftText
        
        # Filter out invalid lines and renumber for SRT compliance
        output_lines = []
        start_number = data.start_line_number or 1
        line_number = start_number
        
        for line in data.lines:
            if line.text and line.start is not None and line.end is not None:
                output_lines.append(SubtitleLine.Construct(
                    line_number, line.start, line.end, line.text, line.metadata
                ))
                line_number += 1

        # Log a warning if any lines had no text or start time
        if len(output_lines) < len(data.lines):
            num_invalid = len([line for line in data.lines if line.start is None])
            if num_invalid:
                logging.warning(_("{} lines were invalid and were not written to the output file").format(num_invalid))

            num_empty = len([line for line in data.lines if not line.text])
            if num_empty:
                logging.warning(_("{} lines were empty and were not written to the output file").format(num_empty))

        # Add RTL markers if required
        if data.metadata.get('add_rtl_markers'):
            for line in output_lines:
                if line.text and IsRightToLeftText(line.text) and not line.text.startswith("\u202b"):
                    line.text = f"\u202b{line.text}\u202c"
        

        srt_items = []
        for line in output_lines:
            proprietary = line.metadata.get('proprietary', '')
            
            srt_item = srt.Subtitle(
                index=line.number,
                start=line.start,
                end=line.end,
                content=line.text,
                proprietary=proprietary
            )
            srt_items.append(srt_item)
        
        return srt.compose(srt_items, reindex=False)

    def _parse_srt_items(self, source) -> Iterator[SubtitleLine]:
        """
        Internal helper to parse SRT items from a file object or string and yield SubtitleLine objects.
        """
        try:
            srt_items = list(srt.parse(source))
            for srt_item in srt_items:
                proprietary = getattr(srt_item, 'proprietary', '')
                metadata = {"proprietary": proprietary} if proprietary else {}

                line = SubtitleLine.Construct(
                    number=srt_item.index,
                    start=srt_item.start,
                    end=srt_item.end,
                    text=srt_item.content,
                    metadata=metadata
                )
                yield line
                
        except UnicodeDecodeError:
            raise  # Re-raise UnicodeDecodeError for fallback handling
        except srt.SRTParseError as e:
            raise SubtitleParseError(_("Failed to parse SRT: {}" ).format(str(e)), e)
        except Exception as e:
            raise SubtitleParseError(_("Unexpected error parsing SRT: {}" ).format(str(e)), e)
    



import logging
from typing import Any
import regex
from datetime import timedelta

from PySubtrans.Helpers.SubtitleHelpers import FindSplitPoint, GetProportionalDuration
from PySubtrans.Helpers.Text import (
    dialog_marker,
    split_sequences,
    break_sequences,
    sentence_end_punctuation,
    BreakLongLine,
    BreakDialogOnOneLine,
    CompileDialogSplitPattern,
    CompileFillerWordsPattern,
    ConvertWhitespaceBlocksToNewlines,
    ConvertWideDashesToStandardDashes,
    EnsureFullWidthPunctuation,
    NormaliseDialogTags,
    RemoveFillerWords
)
from PySubtrans.Options import SettingsType
from PySubtrans.SettingsType import SettingsType
from PySubtrans.SubtitleLine import SubtitleLine

class SubtitleProcessor:
    """
    Helper class to pre-process or post-process subtitles to make them suitable for translation.

    Will split long lines, add line breaks and remove empty lines.
    """
    def __init__(self, settings : SettingsType):
        # Convert plain dict to SettingsType if needed for typed access
        if isinstance(settings, dict) and not isinstance(settings, SettingsType):
            settings = SettingsType(settings)
            
        self.dialog_marker = dialog_marker
        self.split_sequences = split_sequences
        self.break_sequences = break_sequences
        self._compiled_split_sequences = None
        self._compiled_break_sequences = None
        self.forbidden_start_end_pairs = [
            ("<", ">"),
            ("[", "]"),
            ("(", ")"),
            ('"', '"'),
            ("“", "”"),
            ("‘", "’"),
        ]

        self.max_line_duration : timedelta = settings.get_timedelta('max_line_duration', timedelta(seconds=0))
        self.min_line_duration : timedelta = settings.get_timedelta('min_line_duration', timedelta(seconds=0))
        self.merge_line_duration : timedelta = settings.get_timedelta('merge_line_duration', timedelta(seconds=0))
        self.min_gap : timedelta = settings.get_timedelta('min_gap', timedelta(seconds=0.05))
        self.min_split_chars : int = settings.get_int('min_split_chars') or 4

        self.convert_whitespace_to_linebreak : bool = settings.get_bool('whitespaces_to_newline', False)
        self.break_dialog_on_one_line : bool = settings.get_bool('break_dialog_on_one_line', False)
        self.normalise_dialog_tags : bool = settings.get_bool('normalise_dialog_tags', False)
        self.remove_filler_words : bool = settings.get_bool('remove_filler_words', False)
        self.full_width_punctuation : bool = settings.get_bool('full_width_punctuation', False)
        self.convert_wide_dashes : bool = settings.get_bool('convert_wide_dashes', False)

        self.break_long_lines : bool = settings.get_bool('break_long_lines', False)
        self.max_single_line_length : int = settings.get_int('max_single_line_length') or 40
        self.min_single_line_length : int = settings.get_int('min_single_line_length') or 4

        self.split_dialog_pattern: regex.Pattern[Any]|None = CompileDialogSplitPattern(self.dialog_marker) if self.break_dialog_on_one_line else None

        filler_words = settings.get_list('filler_words', [])
        self.filler_words_pattern: regex.Pattern[Any]|None = CompileFillerWordsPattern(filler_words) if self.remove_filler_words else None

        self.split_by_duration: bool = self.max_line_duration.total_seconds() > 0.0

    def PreprocessSubtitles(self, lines : list[SubtitleLine]) -> list[SubtitleLine]:
        """
        Pre-process subtitles to make them suitable for translation.

        Will split long lines, add line breaks to dialog, and remove empty lines.
        """
        if not lines:
            return []

        processed : list[SubtitleLine] = []
        line_number : int = lines[0].number

        if self.merge_line_duration.total_seconds() > 0.0:
            lines = self._merge_short_lines(lines, self.merge_line_duration)

        for line in lines:
            line.number = line_number

            # Preprocess the line (split dialogs, convert whitespace, etc)
            self._preprocess_line(line)

            if not line.text:
                continue

            needs_split : bool = self.split_by_duration and line.duration > self.max_line_duration

            if needs_split:
                split_lines = self._split_line_by_duration(line)

                processed.extend(split_lines)
                line_number += len(split_lines)

                if len(split_lines) > 1:
                    new_line_text = ''.join([str(l) for l in split_lines])
                    logging.debug(f"Split line {line.number} into {len(split_lines)} parts:\n{str(line)}-->\n{new_line_text}")
                else:
                    logging.debug(f"Failed to split line {line.number}:\n{str(line)}")
            else:
                processed.append(line)
                line_number += 1

        return processed

    def PostprocessSubtitles(self, lines : list[SubtitleLine]) -> list[SubtitleLine]:
        """
        Post-process lines after translation
        """
        if not lines:
            return []

        processed : list[SubtitleLine] = []

        for line in lines:
            processed_line = self._postprocess_line(line)
            if processed_line:
                processed.append(processed_line)

        # TODO: fix minimum durations
        # TODO: fix overlapping start/end times (or merge the lines?)

        return processed

    def _preprocess_line(self, line : SubtitleLine):
        """
        Split dialogs onto separate lines.
        Adjust line breaks to split at punctuation weighted by centrality
        """
        text : str = line.text.strip() if line.text else ""
        if not text:
            return

        # Convert whitespace blocks to newlines
        if self.convert_whitespace_to_linebreak:
            text = ConvertWhitespaceBlocksToNewlines(text)

        # Convert wide dashes to standard dashes
        if self.convert_wide_dashes:
            text = ConvertWideDashesToStandardDashes(text)

        # Ensure full-width punctuation is used in Asian languages
        if self.full_width_punctuation:
            text = EnsureFullWidthPunctuation(text)

        # Remove filler words
        if self.remove_filler_words and self.filler_words_pattern:
            text = RemoveFillerWords(text, self.filler_words_pattern)

        # If the subtitle is a single line, see if it should have line breaks added
        if self.break_dialog_on_one_line and self.split_dialog_pattern:
            text = BreakDialogOnOneLine(text, self.split_dialog_pattern)

        # If the subtitle has multiple lines, make sure dialog markers match
        if self.normalise_dialog_tags:
            text = NormaliseDialogTags(text, self.dialog_marker)

        if text != line.text:
            logging.debug(f"Preprocessed line {line.number}:\n{line.text}\n-->\n{text}")
            line.text = text

    def _postprocess_line(self, line : SubtitleLine):
        """
        Split dialogs onto separate lines.
        Normalise dialog markers.
        Add line breaks to long lines.
        """
        text = line.text.strip() if line.text else ""

        if not text:
            return line

        if self.remove_filler_words and self.filler_words_pattern:
            text = RemoveFillerWords(text, self.filler_words_pattern)

        if self.convert_wide_dashes:
            text = ConvertWideDashesToStandardDashes(text)

        if self.break_dialog_on_one_line and self.split_dialog_pattern:
            text = BreakDialogOnOneLine(text, self.split_dialog_pattern)

        if self.normalise_dialog_tags:
            text = NormaliseDialogTags(text, self.dialog_marker)

        if self.full_width_punctuation:
            text = EnsureFullWidthPunctuation(text)

        if self.break_long_lines:
            text = self._break_long_lines(text)

        if text == line.text:
            return line

        logging.debug(f"Postprocessed line {line.number}:\n{line.text}\n-->\n{text}")
        processed_line = SubtitleLine.Construct(line.number, line.start, line.end, text, line.metadata)
        return processed_line

    def _break_long_lines(self, text : str) -> str:
        """
        Add line breaks to long lines
        """
        if self._compiled_break_sequences is None:
            self._compile_break_sequences()

        max_length = self.max_single_line_length
        min_length = self.min_single_line_length
        break_sequences = self._compiled_break_sequences
        if break_sequences:
            text = BreakLongLine(text, max_line_length=max_length, min_line_length=min_length, break_sequences=break_sequences)
        return text

    def _split_line_by_duration(self, line: SubtitleLine) -> list[SubtitleLine]:
        """
        Recursively split a line into smaller lines based on the duration of the text,
        by choosing a split point from the defined sequences weighted towards the middle.
        """
        result : list[SubtitleLine] = []
        stack : list[SubtitleLine] = [line]

        if self._compiled_split_sequences is None:
            self._compile_split_sequences()
            if not self._compiled_split_sequences:
                raise ValueError("No split sequences defined for splitting lines by duration.")

        while stack:
            current_line = stack.pop()
            if not current_line or not current_line.text or not current_line.start or not current_line.end:
                continue

            if current_line.duration <= self.max_line_duration or len(current_line.text) < self.min_split_chars * 2:
                result.append(current_line)
                continue

            if (current_line.text[0], current_line.text[-1]) in self.forbidden_start_end_pairs:
                result.append(current_line)
                continue

            split_point = FindSplitPoint(current_line, self._compiled_split_sequences, min_duration=self.min_line_duration, min_split_chars=self.min_split_chars)
            if split_point is None:
                result.append(current_line)
                continue

            split_text : str = current_line.text[split_point:].strip()
            split_duration : timedelta = GetProportionalDuration(current_line, len(split_text), self.min_line_duration)
            split_start : timedelta = current_line.end - split_duration
            split_end : timedelta = current_line.end

            new_line = SubtitleLine.Construct(current_line.number, current_line.start, split_start - self.min_gap, current_line.text[:split_point], current_line.metadata)
            split_line = SubtitleLine.Construct(current_line.number, split_start, split_end, split_text, current_line.metadata)

            stack.extend([split_line, new_line])

        for i, result_line in enumerate(result, start=0):
            result_line.number = line.number + i

        return result

    def _merge_short_lines(self, lines : list[SubtitleLine], short_duration : timedelta) -> list[SubtitleLine]:
        """
        Merge lines with very short durations into the previous line
        """
        if not lines:
            return []

        merged_lines : list[SubtitleLine] = []
        current_line : SubtitleLine = lines[0]

        for line in lines[1:]:
            if not current_line.text_normalized:
                current_line = line
                continue

            if line.duration < short_duration:
                # If the line ends with a sentence-ending punctuation mark, assume different speakers (questionable logic)
                if current_line.text_normalized[-1] in sentence_end_punctuation:
                    current_line.text = f"{dialog_marker}{current_line.text}\n{dialog_marker}{line.text}"
                else:
                    current_line.text = f"{current_line.text}\n{line.text}"

                current_line.end = line.end
            else:
                merged_lines.append(current_line)
                current_line = line

        merged_lines.append(current_line)
        return merged_lines

    def _compile_split_sequences(self):
        self._compiled_split_sequences = [regex.compile(seq) for seq in self.split_sequences]

    def _compile_break_sequences(self):
        self._compiled_break_sequences = [regex.compile(seq) for seq in self.break_sequences]





import logging
import os
from typing import Any

from PySubtrans.Instructions import Instructions, DEFAULT_TASK_TYPE, default_instructions, default_retry_instructions
from PySubtrans.Helpers.Localization import _
from PySubtrans.Helpers.Resources import GetResourcePath, config_dir

linesep = '\n'

def LoadInstructions(instruction_file : str) -> Instructions:
    """
    Load instruction file with flexible path resolution:
    1. Check if it's an absolute path and exists
    2. Check if it's relative to current directory
    3. Check LLM-Subtrans user directory if exists, otherwise load from resources
    """
    # Check absolute path
    if os.path.isabs(instruction_file) and os.path.exists(instruction_file):
        return LoadInstructionsFile(instruction_file)

    # Check relative to current directory
    if os.path.exists(instruction_file):
        return LoadInstructionsFile(os.path.abspath(instruction_file))

    # Check whether the file exists in the LLM-Subtrans user directory
    user_path = GetInstructionsUserPath(instruction_file)
    if os.path.exists(user_path):
        return LoadInstructionsFile(user_path)

    # Finally, try loading from package resources
    return LoadInstructionsResource(instruction_file)

def LoadInstructionsFile(filepath : str) -> Instructions:
    """
    Load instructions from a text file.
    """
    if not os.path.exists(filepath):
        raise ValueError(f"Instruction file not found: {filepath}")

    instructions = Instructions({})

    with open(filepath, "r", encoding="utf-8") as f:
        lines = [l.strip() for l in f.readlines()]

    if not lines:
        return instructions

    if not lines[0].startswith('###'):
        logging.info(f"Loading legacy instruction file: {filepath}")
        file_instructions, file_retry_instructions = LoadLegacyInstructions(lines)
        if file_instructions:
            instructions.instructions = file_instructions
            instructions.retry_instructions = file_retry_instructions or default_retry_instructions
            instructions.instruction_file = os.path.basename(filepath)
        return instructions

    sections : dict[str, list[str]] = {}
    section_name : str|None = None
    for line in lines:
        if line.startswith('###'):
            section_name = line[3:].strip()
            sections[section_name] = []
        elif section_name is not None and (line.strip() or sections[section_name]):
            sections[section_name].append(line)

    instructions.prompt = linesep.join(sections.get('prompt', [])).strip()
    instructions.instructions = linesep.join(sections.get('instructions', [])).strip()
    instructions.retry_instructions = linesep.join(sections.get('retry_instructions', [])).strip() or default_retry_instructions
    instructions.instruction_file = os.path.basename(filepath)
    instructions.target_language = ''.join(sections.get('target_language', [])) if 'target_language' in sections else None
    instructions.task_type = ''.join(sections.get('task_type', [])) if 'task_type' in sections else DEFAULT_TASK_TYPE

    if not instructions.prompt or not instructions.instructions:
        raise ValueError("Invalid instruction file")

    return instructions


def SaveInstructions(instructions : Instructions, filepath : str) -> None:
    """
    Save instructions to a text file.
    """
    if filepath:
        # Make sure the file has a .txt extension
        if not filepath.endswith('.txt'):
            filepath += '.txt'

        # Write each section to the file with a header
        with open(filepath, "w", encoding="utf-8", newline='') as f:
            f.write("### prompt\n")
            f.write(instructions.prompt or default_instructions)
            if instructions.task_type != DEFAULT_TASK_TYPE:
                f.write("\n\n### task_type\n{}".format(instructions.task_type))
            f.write("\n\n### instructions\n")
            f.write(instructions.instructions or default_instructions)
            f.write("\n\n### retry_instructions\n")
            f.write(instructions.retry_instructions or default_retry_instructions)
            f.write("\n")

        instructions.instruction_file = os.path.basename(filepath)


def LoadLegacyInstructions(lines : list[str]) -> tuple[str|None, str|None]:
    """
    Retry instructions can be added to the file after a line of at least 3 # characters.
    """
    if lines:
        for idx, item in enumerate(lines):
            if len(item) >= 3 and all(c == '#' for c in item):
                return linesep.join(lines[:idx]), linesep.join(lines[idx + 1:])

        return linesep.join(lines), None

    return None, None


def GetInstructionsResourcePath(instructions_file : str|None = None) -> str:
    """
    Get the path for an instructions file (or the directory that contains them).
    """
    if not instructions_file:
        return GetResourcePath("instructions")

    return GetResourcePath("instructions", instructions_file)


def GetInstructionsResourceFiles() -> list[str]:
    """
    Get a list of instruction files in the instructions directory.
    """
    instruction_path = GetInstructionsResourcePath()
    logging.debug(f"Looking for instruction files in {instruction_path}")
    files = os.listdir(instruction_path)
    return [ file for file in files if file.lower().startswith("instructions") ]


def LoadInstructionsResource(resource_name : str) -> Instructions:
    """
    Load instructions from a file in the project/package.
    """
    filepath = GetInstructionsResourcePath(resource_name)
    logging.debug(f"Loading instructions from {filepath}")
    return LoadInstructionsFile(filepath)


def GetInstructionsUserPath(instructions_file : str|None = None) -> str:
    """
    Get the path for an instructions file (or the directory that contains them).
    """
    instructions_dir = os.path.join(config_dir, "instructions")
    return os.path.join(instructions_dir, instructions_file) if instructions_file else instructions_dir


def GetInstructionsUserFiles() -> list[str]:
    """
    Get a list of instruction files in the user directory.
    """
    instructions_dir = GetInstructionsUserPath()
    logging.debug(f"Looking for instruction files in {instructions_dir}")
    if not os.path.exists(instructions_dir):
        os.makedirs(instructions_dir)

    files = os.listdir(instructions_dir)
    return [ file for file in files if file.lower().endswith(".txt") ]


def GetInstructionsFiles() -> list[str]:
    """
    Get a list of instruction files in the user and resource directories.
    """
    default_instructions = GetInstructionsResourceFiles()
    user_instructions = GetInstructionsUserFiles()

    instructions_map = { os.path.basename(file).lower(): file for file in default_instructions }
    instructions_map.update({ os.path.basename(file).lower(): file for file in user_instructions })

    # Sort 'instructions.txt' to the top of the list followed by other names case-insensitive
    return sorted(list(instructions_map.values()), key=lambda x: (x.lower() != 'instructions.txt', x.lower()))



from typing import Any

from PySubtrans.Helpers.Localization import _

class SubtitleError(Exception):
    def __init__(self, message : str|None = None, error : Exception|None = None):
        super().__init__(message)
        self.error = error
        self.message = message

    def __str__(self) -> str:
        if self.error:
            return str(self.error)
        elif self.message:
            return self.message
        return super().__str__()

class NoProviderError(SubtitleError):
    def __init__(self):
        super().__init__(_("Provider not specified in options"))

class ProviderError(SubtitleError):
    def __init__(self, message : str|None = None, provider : Any = None):
        super().__init__(message)
        self.provider = provider
 
class ProviderConfigurationError(ProviderError):
    def __init__(self, message : str, provider : Any, error : Exception|None = None):
        super().__init__(message, provider)
        self.error = error

class TranslationError(SubtitleError):
    def __init__(self, message : str, translation : Any = None, error : Exception|None = None):
        super().__init__(message)
        self.translation = translation
        self.error = error

class TranslationAbortedError(TranslationError):
    def __init__(self):
        super().__init__(_("Translation aborted"))

class TranslationImpossibleError(TranslationError):
    """ No chance of retry succeeding """
    def __init__(self, message : str, error : Exception|None = None):
        super().__init__(message, error)

class TranslationResponseError(TranslationError):
    def __init__(self, message : str, response : Any):
        super().__init__(message)
        self.response = response

class NoTranslationError(TranslationError):
    def __init__(self, message : str, translation : str|None = None):
        super().__init__(message=message, translation=translation)

class TranslationValidationError(TranslationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, translation)
        self.lines = lines or []

class UntranslatedLinesError(TranslationValidationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, lines=lines, translation=translation)

class UnmatchedLinesError(TranslationValidationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, lines=lines, translation=translation)

class EmptyLinesError(TranslationValidationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, lines=lines, translation=translation)

class TooManyNewlinesError(TranslationValidationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, lines=lines, translation=translation)

class LineTooLongError(TranslationValidationError):
    def __init__(self, message : str, lines : list[Any]|None = None, translation : Any|None = None):
        super().__init__(message, lines=lines, translation=translation)

class SubtitleParseError(SubtitleError):
    """Error raised when subtitle file cannot be parsed."""
    def __init__(self, message : str, error : Exception|None = None):
        super().__init__(message, error)
 



from typing import Any
import unicodedata
import regex
import unicodedata
from collections import Counter

common_punctuation = r"[.,!?;:…¡¿]"
sentence_end_punctuation = r"[.!?…？！。﹑]"

dialog_marker = "- "
emdash = "—"

standard_filler_words = "um,umm,uh,uhh,er,err,ah,ahh,oh,eh,hm,hmm,hmmm,huh,ha,mmm,ow,oww"

whitespace_and_punctuation_pattern = regex.compile(r'[\p{P}\p{Z}\p{C}]')

priority_break_sequences = [
    regex.escape(dialog_marker),  # Dialog marker
    r"(?=\([^)]*\)|\[[^\]]*\])",  # Look ahead to find a complete parenthetical or bracketed block to split before
    r"(?=\"[^\"]*\")",  # Look ahead to find a complete block within double quotation marks
    r"(?=<([ib])>[^<]*</\1>)",  # Look ahead to find a block in italics or bold
]

break_sequences = priority_break_sequences + [
    r"[.!?](\s|\")",  # End of sentence punctuation like '!', '?', possibly at the end of a quote
    r"[？！。…，、﹑]", # Full-width punctuation (does not need to be followed by whitespace)
    r"[,](\s|\")",  # Commas followed by whitespace or quote
    r"[:;]\s+",  # Colon and semicolon
    r"[–—]+\s+",  # Dashes
    r"\s+",  # Whitespace
]

split_sequences = [
    r"\n",  # Newline has the highest priority
    regex.escape(dialog_marker),  # Dialog marker
    r"(?=\([^)]*\)|\[[^\]]*\])",  # Look ahead to find a complete parenthetical or bracketed block to split before
    r"(?=\"[^\"]*\")",  # Look ahead to find a complete block within double quotation marks
    r"(?=<([ib])>[^<]*</\1>)",  # Look ahead to find a block in italics or bold
    r"[.!?](\s|\")",  # End of sentence punctuation like '!', '?', possibly at the end of a quote
    r"[？！。…，、﹑]", # Full-width punctuation (does not need to be followed by whitespace)
    r"[,](\s|\")",  # Commas followed by whitespace or quote
    r"[:;；：]\s+",  # Colon and semicolon
    r"[–—]+\s+",  # Dashes
    r" {3,}"  # Three or more spaces
]

# Dictionary mapping half-width punctuation to full-width
fullwidth_punctuation_map = {
    ',': '，',
    '.': '。',
    ';': '；',
    ':': '：',
    '?': '？',
    '!': '！'
}

# Regex to find half-width punctuation adjacent to Asian script characters
fullwidth_pattern = r'(?<=[\p{Script=Han}\p{Script=Hangul}\p{Script=Hiragana}\p{Script=Katakana}])(?P<punct>[,.;:?!\-])(?=[\p{Script=Han}\p{Script=Hangul}\p{Script=Hiragana}\p{Script=Katakana}])'

def RemoveWhitespaceAndPunctuation(string) -> str:
    """
    Remove all whitespace and punctuation from a string
    """
    # Matches any punctuation, separator, or other Unicode character
    stripped = whitespace_and_punctuation_pattern.sub('', string)

    # Normalize Unicode characters to their canonical forms
    normalized = unicodedata.normalize('NFC', stripped)

    return normalized

def IsTextContentEqual(string1 : str|None, string2 : str|None) -> bool:
    """
    Compare two strings for equality, ignoring whitespace and punctuation
    """
    if string1 and string2:
        stripped1 = RemoveWhitespaceAndPunctuation(string1)
        stripped2 = RemoveWhitespaceAndPunctuation(string2)
        return stripped1 == stripped2

    return string1 == string2

def Linearise(lines : str|list[str]) -> str:
    """
    Ensure that the input is a single string
    """
    if not isinstance(lines, list):
        lines = str(lines).split("\n")

    lines = [ str(line).strip() for line in lines ]
    return " | ".join(lines)

def ConvertWhitespaceBlocksToNewlines(text : str) -> str:
    """
    Convert blocks of 3 or more spaces or chinese commas to newlines, unless the text contains newlines already
    """
    if text and '\n' not in text:
        text = regex.sub(r' {3,}|\，\s*', '\n', text)

    return text

def ConvertWideDashesToStandardDashes(text : str) -> str:
    """
    """
    text = regex.sub(r'\s*—+\s*', ' - ', text)
    return text

def EnsureFullWidthPunctuation(text: str) -> str:
    """
    Ensure full-width punctuation is used in East Asian languages by replacing half-width
    punctuation with full-width equivalents only when directly adjacent to Asian script characters.
    """
    # Function to replace each punctuation mark found with its full-width counterpart
    def replace(match):
        punctuation = match.group('punct')
        return fullwidth_punctuation_map[punctuation]

    # Replace all occurrences of half-width punctuation in the text
    return regex.sub(fullwidth_pattern, replace, text)


def CompileDialogSplitPattern(dialog_marker):
    """
    Compile a regex pattern to split lines at dialog markers
    """
    escaped_marker = regex.escape(dialog_marker)
    re_split = r"(?<=[^a-zA-Z0-9\s])\s*(?=" + escaped_marker + ")"
    return regex.compile(re_split)

def BreakDialogOnOneLine(text : str, dialog_marker : str|regex.Pattern) -> str:
    """
    Break dialog into separate lines
    """
    # Split line at dialog markers following any non-alphanumeric character and whitespace
    # This should catch the majority of genuine dialog markers and few other uses of a dash
    # Uses a look-behind followed by a look-ahead so that the split point is not consumed
    if not isinstance(dialog_marker, regex.Pattern):
        dialog_marker = CompileDialogSplitPattern(dialog_marker)

    line_parts = dialog_marker.split(text)

    if len(line_parts) > 1:
        text = '\n'.join([part.strip() for part in line_parts])

    return text

def NormaliseDialogTags(text : str, dialog_marker : str) -> str:
    """
    Make sure dialog markers are consistent across lines
    """
    if not dialog_marker in text:
        return text

    line_parts = text.split('\n')

    # If a single line starts with a dialog marker, remove it
    if len(line_parts) == 1 and text.startswith(dialog_marker):
        return text[len(dialog_marker):].strip()

    # If any of the line parts starts with a dialog marker, they all should
    if any(part.startswith(dialog_marker) for part in line_parts):
        line_parts = [part if part.startswith(dialog_marker) else dialog_marker + part for part in line_parts]
        text = '\n'.join(part.strip() for part in line_parts)

    return text

def FindBreakPoint(text : str, break_sequences: list[regex.Pattern], max_line_length : int, min_line_length : int) -> int|None:
    """
    Find the optimal break point for a long line
    """
    line_length = len(text)
    start_index = min_line_length
    end_index = line_length - min_line_length
    if end_index <= start_index:
        return None

    middle_index = line_length // 2

    min_break = min(line_length - max_line_length, max_line_length)
    min_break = max(min_break, min_line_length)

    for priority, seq in enumerate(break_sequences, start=1):
        matches = list(seq.finditer(text))
        if not matches:
            continue

        # Find the match that is closest to the middle of the text
        best_match = min(matches, key=lambda m: abs(m.end() - middle_index))
        split_index = best_match.end()
        if split_index < start_index or split_index > end_index:
            continue

        # Don't break if it would result in a line longer than the maximum or shorter than the minimum, if avoidable
        if priority > len(priority_break_sequences) and split_index < min_break:
            continue

        return split_index

    return None

def BreakLongLine(text : str, max_line_length : int, min_line_length : int, break_sequences: list[regex.Pattern]) -> str:
    """
    Add line breaks to long single lines
    """
    length = len(text)
    if length <= max_line_length:
        return text

    if '\n' in text:
        return text

    break_index = FindBreakPoint(text, break_sequences, max_line_length, min_line_length)
    if break_index:
        text = text[:break_index].strip() + '\n' + text[break_index:].strip()

    return text

def LimitTextLength(text : str, max_length : int) -> str:
    """
    Limit the length of a text string to a maximum number of characters, cutting at the nearest sentence end or whitespace
    """
    text = text.strip()

    if len(text) <= max_length:
        return text

    pattern = r'\.|\?|!'
    matches = [(m.start(), m.group()) for m in regex.finditer(pattern, text)]

    # Find the closest match to the max length, if any
    for position, match in reversed(matches): # type: ignore
        if position <= max_length:
            return text[:position + 1]

    # If no sentence end is found within the limit, cut at the nearest whitespace
    nearest_space = text.rfind(' ', 0, max_length)
    if nearest_space != -1:
        return text[:nearest_space] + '...'
    else:
        # As a last resort, cut directly at the max length
        return text[:max_length] + '...'

def CompileFillerWordsPattern(filler_words: str|list[str]) -> regex.Pattern[Any]|None:
    """
    Compile a regex pattern to match any provided filler word, assuming they are
    followed by mandatory punctuation and possibly preceded by punctuation.
    """
    if isinstance(filler_words, str):
        filler_words = [i.strip() for i in filler_words.split(',') if i.strip()]

    if not filler_words:
        return None

    filler_pattern = '|'.join(regex.escape(i) for i in filler_words if i)
    filler_words_pattern = rf"(^|[,¡¿]?\s+)({filler_pattern})({common_punctuation}+(\s+|$))"

    return regex.compile(filler_words_pattern, flags=regex.IGNORECASE|regex.MULTILINE)

def RemoveFillerWords(text: str, fillerWords: str|list[str]|regex.Pattern[Any]) -> str:
    """
    Remove filler words from a text string, adjusting capitalization based on the capitalization of the filler word.
    """
    fillerPatterns = fillerWords if isinstance(fillerWords, regex.Pattern) else CompileFillerWordsPattern(fillerWords)

    if fillerPatterns is None:
        return text

    output = []
    last_index = 0
    capitalise_first_letter = False

    def _append_previous_section(output : list[str], text : str, previous_start : int, previous_end : int, capitalise_first_letter : bool):
        if previous_end > previous_start:
            if capitalise_first_letter:
                first_letter = text[previous_start].upper()
                next_index = previous_start + 1
                previous_section = first_letter if next_index == previous_end else first_letter + text[next_index:previous_end]
                output.append(previous_section)
            else:
                output.append(text[previous_start:previous_end])

    for match in fillerPatterns.finditer(text):
        start, end = match.span()

        # Output any text before the match
        _append_previous_section(output, text, last_index, start, capitalise_first_letter)

        # Capitalize the first letter of the next section if the filler word was capitalized
        first_alpha = next((char for char in text[start:end] if char.isalpha()), None)
        capitalise_first_letter = first_alpha is not None and first_alpha.isupper()

        last_index = end

    # Append any remaining text after the last match
    if last_index < len(text):
        _append_previous_section(output, text, last_index, len(text), capitalise_first_letter)

    # Join the output sections into a single string
    return ' '.join(output)

def ContainsTags(text : str) -> bool:
    """
    Check if a line contains any html-like tags (<i>, <b>, etc.)
    """
    return regex.search(r"<[^>]+>", text) is not None

def ExtractTag(tagname : str, text : str) -> tuple[str, str|None]:
    """
    Look for an xml-like tag in the input text, and extract the contents.
    """
    if not tagname:
        return text, None

    open_tag = f"<{tagname}>"
    close_tag = f"</{tagname}>"
    empty_tag = f"<{tagname}/>"

    text = text.replace(empty_tag, '')

    end_index = text.rfind(close_tag)

    if end_index == -1:
        return text.strip(), None

    start_index = text.rfind(open_tag, 0, end_index)
    if start_index == -1:
        raise ValueError(f"Malformed {tagname} tags in {text}")

    tag = text[start_index + len(open_tag):end_index].strip()
    text_before = text[:start_index].strip()
    text_after = text[end_index + len(close_tag):].strip()
    text = '\n'.join([text_before, text_after]).strip()

    return text, tag

def ExtractTagList(tagname, text):
    """
    Look for an xml-like tag in the input text, and extract the contents as a comma or newline separated list.
    """
    if text is not None:
        text, tag = ExtractTag(tagname, text)
        tag_list = [ item.strip() for item in regex.split("[\n,]", tag) ] if tag else []
        return text, tag_list
    return text, []

def SanitiseSummary(summary : str, movie_name : str|None = None, max_summary_length : int|None = None):
    """
    Remove trivial parts of summary text and limit the length if required
    """
    if not summary:
        return None

    summary = summary.replace("Summary of the batch", "")
    summary = summary.replace("Summary of the scene", "")
    summary = regex.sub(r'^(?:(?:Scene|Batch)[\s\d:\-]*)+', '', summary, flags=regex.IGNORECASE)

    if movie_name:
        # Remove movie name and any connectors (-,: or whitespace)
        summary = regex.sub(r'^' + regex.escape(movie_name) + r'\s*[:\-\s]*', '', summary)

    summary = regex.sub(r'^[\s\d:\-]*', '', summary, flags=regex.IGNORECASE)

    summary = summary.strip()

    if max_summary_length:
        summary = LimitTextLength(summary, max_summary_length)

    return summary or None

def IsRightToLeftText(text: str) -> bool:
    """
    Check if text is predominantly RTL using Unicode bidirectional properties
    """
    if not text:
        return False
    count = Counter(unicodedata.bidirectional(c) for c in text if not c.isspace())
    rtl_count = sum(count[d] for d in ['R', 'AL', 'RLE', 'RLI'])
    ltr_count = sum(count[d] for d in ['L', 'LRE', 'LRI'])
    return rtl_count > ltr_count



import pysubs2
import pysubs2.time
import regex
from datetime import timedelta
from typing import TextIO

from PySubtrans.Helpers.Color import Color
from PySubtrans.SubtitleFileHandler import (
    SubtitleFileHandler,
    default_encoding,
    fallback_encoding,
)
from PySubtrans.SubtitleLine import SubtitleLine
from PySubtrans.SubtitleData import SubtitleData
from PySubtrans.SubtitleError import SubtitleParseError
from PySubtrans.Helpers.Localization import _

# Precompiled regex patterns for performance
_START_TAGS_PATTERN = regex.compile(r'^(\{[^}]*\})+')
_TAG_BLOCK_PATTERN = regex.compile(r'\{[^}]+\}')
_STANDALONE_BASIC_TAG_PATTERN = regex.compile(r'^\{\\(?:[ibs][01]|u[01]?)\}$')
# Matches SSA basic formatting tags: italic (\i1, \i0), bold (\b1, \b0), strikeout (\s1, \s0), underline (\u1, \u0)
_BASIC_TAG_PATTERN = regex.compile(r'\\(?:[ibs][01]|u[01]?)')

# Precompiled SSA to HTML conversion patterns
_SSA_TO_HTML_PATTERNS = [
    (regex.compile(r'{\\i1}'), '<i>'),
    (regex.compile(r'{\\i0}'), '</i>'),
    (regex.compile(r'{\\b1}'), '<b>'),
    (regex.compile(r'{\\b0}'), '</b>'),
    (regex.compile(r'{\\u1}'), '<u>'),
    (regex.compile(r'{\\u0}'), '</u>'),
    (regex.compile(r'{\\s1}'), '<s>'),
    (regex.compile(r'{\\s0}'), '</s>')
]

# Precompiled HTML to SSA conversion patterns
_HTML_TO_SSA_PATTERNS = [
    (regex.compile(r'<i>'), r'{\\i1}'),
    (regex.compile(r'</i>'), r'{\\i0}'),
    (regex.compile(r'<b>'), r'{\\b1}'),
    (regex.compile(r'</b>'), r'{\\b0}'),
    (regex.compile(r'<u>'), r'{\\u1}'),
    (regex.compile(r'</u>'), r'{\\u0}'),
    (regex.compile(r'<s>'), r'{\\s1}'),
    (regex.compile(r'</s>'), r'{\\s0}')
]


class SSAFileHandler(SubtitleFileHandler):
    """
    File handler for Advanced SubStation Alpha (SSA/ASS) subtitle format using pysubs2 library.

    Supports reading and writing SSA/ASS with file- and line-level metadata.
    """
    
    SUPPORTED_EXTENSIONS = {'.ass': 10, '.ssa': 10}

    def load_file(self, path: str) -> SubtitleData:
        # TEST: Use .ssa extension as format hint to detect Marked lines ... this causes more problems than it fixes, let's trust pysubs2
        # format_hint = 'ssa' if path.lower().endswith('.ssa') else None
        format_hint = None
        
        try:
            subs: pysubs2.SSAFile = pysubs2.SSAFile.load(path, encoding=default_encoding, format_=format_hint)
            return self._parse_subs(subs)
        except UnicodeDecodeError:
            subs: pysubs2.SSAFile = pysubs2.SSAFile.load(path, encoding=fallback_encoding, format_=format_hint, newline='')
            return self._parse_subs(subs)
        except Exception as e:
            raise e if isinstance(e, SubtitleParseError) else SubtitleParseError(_("Failed to parse subtitles: {}" ).format(str(e)), e)
    
    def parse_file(self, file_obj: TextIO) -> SubtitleData:
        """
        Parse file content and return SubtitleData with lines and metadata.
        """
        try:
            subs : pysubs2.SSAFile = pysubs2.SSAFile.from_file(file_obj)
            return self._parse_subs(subs)
                
        except UnicodeDecodeError:
            raise  # Re-raise UnicodeDecodeError for fallback handling
        except Exception as e:
            raise SubtitleParseError(_("Failed to parse file: {}").format(str(e)), e)

    def parse_string(self, content: str) -> SubtitleData:
        """
        Parse string content and return SubtitleData with lines and metadata.
        """
        try:
            subs = pysubs2.SSAFile.from_string(content)
            return self._parse_subs(subs)
                
        except Exception as e:
            raise SubtitleParseError(_("Failed to parse content: {}").format(str(e)), e)
    
    def compose(self, data: SubtitleData) -> str:
        """
        Compose subtitle lines into SSA/ASS format string using metadata.
        
        Args:
            data: SubtitleData containing lines and file metadata
            
        Returns:
            str: formatted subtitle content
        """
        subs : pysubs2.SSAFile = pysubs2.SSAFile()
        subs.info["TranslatedBy"] = "LLM-Subtrans"

        self._build_metadata(subs, data.metadata)

        # Restore original detected format
        file_format = data.metadata.get('pysubs2_format', 'ass')

        # Convert SubtitleLines to pysubs2 format
        for line in data.lines:
            if line.text and line.start is not None and line.end is not None:
                pysubs2_line = self._subtitle_line_to_pysubs2(line)
                subs.append(pysubs2_line)
        
        return subs.to_string(file_format)
    
    def _parse_subs(self, subs : pysubs2.SSAFile):
        """
        Convert pysubs2 subtitles to SubtitleLines, adding an index
        """
        try:
            file_format : str = getattr(subs, "format", "ass")

            lines : list[SubtitleLine] = []
            for index, line in enumerate(subs, 1):
                lines.append(self._pysubs2_to_subtitle_line(line, index))

            # Extract serializable metadata
            metadata = self._parse_metadata(subs, file_format)

            detected_format : str = pysubs2.formats.get_file_extension(file_format)

            return SubtitleData(lines=lines, metadata=metadata, detected_format=detected_format)

        except Exception as e:
            raise SubtitleParseError(_("Failed to parse subtitles: {}" ).format(str(e)), e)
        
    def _pysubs2_to_subtitle_line(self, pysubs2_line: pysubs2.SSAEvent, index: int) -> SubtitleLine:
        """Convert pysubs2 SSAEvent to SubtitleLine with metadata preservation."""
        
        start = timedelta(milliseconds=pysubs2_line.start)
        end = timedelta(milliseconds=pysubs2_line.end)
        
        # Extract text content using .text to preserve inline formatting,
        # then convert SSA tags to HTML for SRT and GUI compatibility
        text = self._ssa_to_html(pysubs2_line.text)
        
        metadata = {
            'style': pysubs2_line.style,
            'layer': pysubs2_line.layer,
            'name': pysubs2_line.name,
            'margin_l': pysubs2_line.marginl,
            'margin_r': pysubs2_line.marginr,
            'margin_v': pysubs2_line.marginv,
            'effect': pysubs2_line.effect,
            'type': pysubs2_line.type,
            'marked': getattr(pysubs2_line, 'marked', False)
        }
        
        # Extract whole-line SSA override tags and store in metadata
        if pysubs2_line.text:
            extracted_tags = self._extract_whole_line_tags(pysubs2_line.text)
            if extracted_tags:
                metadata.update(extracted_tags)
        
        return SubtitleLine.Construct(
            number=index,
            start=start,
            end=end,
            text=text,
            metadata=metadata
        )
    
    def _subtitle_line_to_pysubs2(self, line: SubtitleLine) -> pysubs2.SSAEvent:
        """
        Convert SubtitleLine back to pysubs2 SSAEvent using preserved metadata.
        """
        event = pysubs2.SSAEvent()
        
        if line.start:
            event.start = pysubs2.time.make_time(s=line.start.total_seconds())
        else:
            event.start = 0
            
        if line.end:
            event.end = pysubs2.time.make_time(s=line.end.total_seconds())
        else:
            event.end = 0
        
        # Convert HTML tags back to SSA tags, then set text directly
        ssa_text = self._html_to_ass(line.text or "")
        
        ssa_text = self._restore_whole_line_tags(ssa_text, line.metadata or {})
        
        event.text = ssa_text
        
        # Restore metadata if available, otherwise use pysubs2 defaults
        if line.metadata:
            event.style = line.metadata.get('style', event.style)
            event.layer = line.metadata.get('layer', event.layer)
            event.name = line.metadata.get('name', event.name)
            event.marginl = line.metadata.get('margin_l', event.marginl)
            event.marginr = line.metadata.get('margin_r', event.marginr)
            event.marginv = line.metadata.get('margin_v', event.marginv)
            event.effect = line.metadata.get('effect', event.effect)
            event.type = line.metadata.get('type', event.type)
            event.marked = line.metadata.get('marked', False)
        
        return event

    def _parse_metadata(self, subs : pysubs2.SSAFile, subtitle_format : str) -> dict:
        """
        Convert pysubs2 metadata to JSON-serializable format.
        Handles Color objects and other pysubs2-specific types.
        """
        # Extract serializable metadata from the pysubs2 file
        metadata = {
            'pysubs2_format': subtitle_format,
            'info': dict(subs.info),  # Script info section
            'aegisub_project': dict(subs.aegisub_project) if hasattr(subs, 'aegisub_project') else {}
        }
        
        styles = {}
        for name, style in subs.styles.items():
            style_dict = style.as_dict()
            
            for field, value in style_dict.items():
                if isinstance(value, pysubs2.Color):
                    style_dict[field] = Color(value.r, value.g, value.b, value.a)
            
            styles[name] = style_dict
        
        metadata['styles'] = styles
        return metadata
    
    def _build_metadata(self, subs : pysubs2.SSAFile, metadata : dict) -> None:
        """
        Restore pysubs2 metadata from JSON-serialized format.
        Converts Color dicts back to pysubs2.Color objects.
        """
        # Restore script info from metadata
        if 'info' in metadata:
            subs.info.update(metadata['info'])

        if 'Title' in metadata:
            subs.info['Title'] = metadata['Title']
        
        if 'Language' in metadata:
            subs.info['Language'] = metadata['Language']

        # Restore styles from metadata
        if 'styles' in metadata:
            # Clear default styles that pysubs2 creates automatically to avoid conflicts
            subs.styles.clear()
            
            for style_name, style_fields in metadata['styles'].items():
                style_data = style_fields.copy()
                
                for field, value in style_data.items():
                    if isinstance(value, Color):
                        style_data[field] = pysubs2.Color(value.r, value.g, value.b, value.a)
                
                style = pysubs2.SSAStyle(**style_data)
                subs.styles[style_name] = style
        
        # Restore aegisub project data if present
        if 'aegisub_project' in metadata and hasattr(subs, 'aegisub_project'):
            subs.aegisub_project.update(metadata['aegisub_project'])
    
    def _extract_whole_line_tags(self, ssa_text: str) -> dict:
        """Extract whole-line SSA override tags from start of line and return as metadata."""
        if not ssa_text:
            return {}
            
        metadata = {}
        
        # Match consecutive SSA tags at the start of the line
        start_tags_match = _START_TAGS_PATTERN.match(ssa_text)
        if start_tags_match:
            tags_section = start_tags_match.group(0)
            
            # Extract only the complex (non-basic formatting) tags
            complex_tags = []
            for tag_match in _TAG_BLOCK_PATTERN.finditer(tags_section):
                tag = tag_match.group(0)
                # If it's a standalone basic formatting tag, skip it entirely
                if _STANDALONE_BASIC_TAG_PATTERN.match(tag):
                    continue
                
                # For composite tags, remove basic formatting but keep the rest
                cleaned_tag = _BASIC_TAG_PATTERN.sub('', tag)

                if cleaned_tag != '{}':
                    complex_tags.append(cleaned_tag)
            
            if complex_tags:
                metadata['override_tags_start'] = ''.join(complex_tags)
        
        return metadata
    
    def _restore_whole_line_tags(self, text: str, metadata: dict) -> str:
        """Restore whole-line SSA tags from metadata."""
        if 'override_tags_start' in metadata:
            return f"{metadata['override_tags_start']}{text}"
        return text
    
    def _ssa_to_html(self, ssa_text: str) -> str:
        """Convert SSA inline formatting tags to HTML tags for GUI display."""
        if not ssa_text:
            return ""
            
        text = ssa_text
        
        # Remove complex whole-line tags at the start (they'll be stored in metadata)
        # But preserve basic formatting tags for HTML conversion
        start_match = _START_TAGS_PATTERN.match(text)
        if start_match:
            tags_section = start_match.group(0)
            remaining_text = text[len(tags_section):]

            # Extract basic formatting tags from both standalone blocks and composite blocks
            basic_tags = []
            for tag_match in _TAG_BLOCK_PATTERN.finditer(tags_section):
                tag_content = tag_match.group(0)
                if _STANDALONE_BASIC_TAG_PATTERN.match(tag_content):
                    basic_tags.append(tag_content)
                else:
                    for basic_tag in _BASIC_TAG_PATTERN.finditer(tag_content):
                        basic_tags.append('{' + basic_tag.group(0) + '}')

            # Rebuild text with only basic formatting tags preserved
            text = ''.join(basic_tags) + remaining_text

        # Extract basic tags from composite blocks anywhere in the line
        rebuilt = []
        last_end = 0
        for tag_match in _TAG_BLOCK_PATTERN.finditer(text):
            block = tag_match.group(0)
            rebuilt.append(text[last_end:tag_match.start()])

            if _STANDALONE_BASIC_TAG_PATTERN.match(block):
                rebuilt.append(block)
            else:
                basic_tags = ['{' + m.group(0) + '}' for m in _BASIC_TAG_PATTERN.finditer(block)]
                cleaned = _BASIC_TAG_PATTERN.sub('', block)
                if cleaned != '{}':
                    rebuilt.append(cleaned)
                rebuilt.extend(basic_tags)

            last_end = tag_match.end()

        rebuilt.append(text[last_end:])
        text = ''.join(rebuilt)
        
        # Convert line breaks:
        # \N (hard line break) -> \n (newline for GUI)
        # \n (soft line break) -> <wbr> (word break opportunity)
        text = text.replace('\\N', '\n')
        text = text.replace('\\n', '<wbr>')
        
        for pattern, replacement in _SSA_TO_HTML_PATTERNS:
            text = pattern.sub(replacement, text)
        
        # For any remaining SSA tags that aren't basic formatting, preserve them
        # This allows translators to see and preserve complex inline formatting
        # (colors, fonts, etc.) that don't have HTML equivalents
        
        return text
    
    def _html_to_ass(self, html_text: str) -> str:
        """Convert HTML tags back to SSA inline formatting tags."""
        if not html_text:
            return ""
            
        text = html_text
        
        # Convert line breaks back:
        # \n (newline from GUI) -> \N (hard line break)
        # <wbr> (word break opportunity) -> \n (soft line break)
        text = text.replace('<wbr>', '\\n')
        text = text.replace('\n', '\\N')
        
        for pattern, replacement in _HTML_TO_SSA_PATTERNS:
            text = pattern.sub(replacement, text)
        
        # Preserve any other HTML that might be part of the dialogue content
        # (e.g., someone translating a movie about HTML)
        
        return text




from PySubtrans.Helpers.Localization import _
from PySubtrans.Providers.Clients.CustomClient import CustomClient
from PySubtrans.SettingsType import SettingsType

class DeepSeekClient(CustomClient):
    """
    Handles chat communication with DeepSeek to request translations using CustomClient logic
    """
    def __init__(self, settings: SettingsType):
        settings['supports_system_messages'] = True
        settings['supports_conversation'] = True
        settings['supports_reasoning'] = True
        settings['supports_streaming'] = True
        settings.setdefault('server_address', settings.get_str('api_base', 'https://api.deepseek.com'))
        settings.setdefault('endpoint', '/v1/chat/completions')
        super().__init__(settings)



from os import linesep
import logging
import threading
from typing import Any

from PySubtrans.Helpers.ContextHelpers import GetBatchContext
from PySubtrans.Helpers.SubtitleHelpers import MergeTranslations
from PySubtrans.Helpers.Localization import _, tr
from PySubtrans.Helpers.Text import Linearise, SanitiseSummary
from PySubtrans.Instructions import DEFAULT_TASK_TYPE, Instructions
from PySubtrans.SettingsType import SettingsType
from PySubtrans.Substitutions import Substitutions
from PySubtrans.SubtitleLine import SubtitleLine
from PySubtrans.SubtitleProcessor import SubtitleProcessor
from PySubtrans.SubtitleValidator import SubtitleValidator
from PySubtrans.Translation import Translation
from PySubtrans.TranslationClient import TranslationClient
from PySubtrans.TranslationParser import TranslationParser
from PySubtrans.Options import Options, SettingsType
from PySubtrans.SubtitleBatch import SubtitleBatch

from PySubtrans.SubtitleError import NoProviderError, NoTranslationError, ProviderError, SubtitleError, TranslationAbortedError, TranslationError, TranslationImpossibleError
from PySubtrans.Helpers import FormatErrorMessages
from PySubtrans.Subtitles import Subtitles
from PySubtrans.SubtitleScene import SubtitleScene, UnbatchScenes
from PySubtrans.TranslationEvents import TranslationEvents
from PySubtrans.TranslationPrompt import TranslationPrompt
from PySubtrans.TranslationProvider import TranslationProvider
from PySubtrans.TranslationRequest import StreamingCallback

class SubtitleTranslator:
    """
    Processes subtitles into scenes and batches and sends them for translation
    """
    def __init__(self, settings: Options, translation_provider: TranslationProvider, resume: bool = False):
        """
        Initialise a SubtitleTranslator with translation options
        """
        self.events = TranslationEvents()
        self.lock = threading.Lock()
        self.aborted : bool = False
        self.errors : list[str|SubtitleError] = []
        self.lines_processed : int = 0

        self.max_lines = settings.get_int('max_lines')
        self.max_history = settings.get_int('max_context_summaries')
        self.stop_on_error = settings.get_bool('stop_on_error')
        self.retry_on_error = settings.get_bool('retry_on_error')
        # self.split_on_error = options.get('autosplit_incomplete')
        self.max_summary_length = settings.get_int('max_summary_length')
        self.retranslate = settings.get_bool('retranslate')
        self.reparse = settings.get_bool('reparse')
        self.preview = settings.get_bool('preview')
        self.resume = resume and (not self.reparse and not self.retranslate)

        settings = Options(settings)

        self.instructions : Instructions = settings.GetInstructions()
        self.task_type : str = self.instructions.task_type or DEFAULT_TASK_TYPE
        self.user_prompt : str = settings.BuildUserPrompt()

        substitutions_mode = settings.get_str('substitution_mode') or Substitutions.Mode.Auto
        substitutions_list = settings.get('substitutions', {})
        if not isinstance(substitutions_list, (dict, list, str)):
            logging.warning(_("Invalid substitutions list, must be a dictionary, list or string"))
            substitutions_list = {}

        self.substitutions = Substitutions(substitutions_list, substitutions_mode)

        self.settings : SettingsType = settings.GetSettings()
        self.settings['instructions'] = self.instructions.instructions
        self.settings['retry_instructions'] = self.instructions.retry_instructions

        logging.debug(f"Translation prompt: {self.user_prompt}")

        self.translation_provider : TranslationProvider = translation_provider

        if not self.translation_provider:
            raise NoProviderError()

        try:
            self.client : TranslationClient = self.translation_provider.GetTranslationClient(self.settings)

        except Exception as e:
            raise ProviderError(_("Unable to create provider client: {error}").format(error=str(e)), translation_provider)

        if not self.client:
            raise ProviderError(_("Unable to create translation client"), translation_provider)

        self.postprocessor = SubtitleProcessor(settings) if settings.get('postprocess_translation') else None
        self.validator = SubtitleValidator(settings)

    def StopTranslating(self):
        self.aborted = True
        self.client.AbortTranslation()

    def TranslateSubtitles(self, subtitles : Subtitles):
        """
        Translate a SubtitleFile
        """
        if not subtitles:
            raise TranslationImpossibleError(_("No subtitles to translate"))

        if subtitles.scenes and self.resume:
            self._emit_info(_("Resuming translation"))

        if not subtitles.scenes:
            raise TranslationImpossibleError(_("Subtitles must be batched before translation"))

        self._emit_info(_("Translating {linecount} lines in {scenecount} scenes").format(linecount=subtitles.linecount, scenecount=subtitles.scenecount))

        self.events.preprocessed.send(self, scenes=subtitles.scenes)

        # Iterate over each subtitle scene and request translation
        for scene in subtitles.scenes:
            if self.aborted:
                break

            if self.max_lines and self.lines_processed >= self.max_lines:
                break

            if self.resume and scene.all_translated:
                self._emit_info(_("Scene {scene} already translated {linecount} lines...").format(scene=scene.number, linecount=scene.linecount))
                continue

            logging.debug(f"Translating scene {scene.number} of {subtitles.scenecount}")
            batch_numbers =[ batch.number for batch in scene.batches if not batch.translated ] if self.resume else None

            self.TranslateScene(subtitles, scene, batch_numbers=batch_numbers)

            if self.errors and self.stop_on_error:
                self._emit_error(_("Failed to translate scene {scene}... stopping translation").format(scene=scene.number))
                return

        if self.aborted:
            self._emit_info(_("Translation aborted"))
            return

        # Linearise the translated scenes
        originals, translations, untranslated = UnbatchScenes(subtitles.scenes)

        if translations:
            self._emit_info(_("Successfully translated {count} lines!").format(count=len(translations)))

        if untranslated and not self.max_lines and not self.preview:
            logging.warning(_("Failed to translate {count} lines:").format(count=len(untranslated)))
            for line in untranslated:
                self._emit_info(_("Untranslated > {number}. {text}").format(number=line.number, text=line.text))

        subtitles.originals = originals
        subtitles.translated = translations

    def TranslateScene(self, subtitles : Subtitles, scene : SubtitleScene, batch_numbers = None, line_numbers = None):
        """
        Send a scene for translation
        """
        try:
            batches = [ batch for batch in scene.batches if batch.number in batch_numbers ] if batch_numbers else scene.batches
            context = {}

            for batch in batches:
                context = GetBatchContext(subtitles, scene.number, batch.number, self.max_history)

                try:
                    self.TranslateBatch(batch, line_numbers, context)

                except TranslationImpossibleError as e:
                    # If batch has any translated lines from streaming, validate them before re-raising
                    if batch.any_translated:
                        self.validator.ValidateBatch(batch)
                    raise

                except TranslationError as e:
                    self._emit_warning(_("Error translating scene {scene} batch {batch}: {error}").format(scene=batch.scene, batch=batch.number, error=str(e)))
                    batch.errors.append(e)

                if self.aborted:
                    return

                # Notify observers the batch was translated
                self.events.batch_translated.send(self, batch=batch)

                if batch.errors:
                    self._emit_warning(_("Errors encountered translating scene {scene} batch {batch}").format(scene=batch.scene, batch=batch.number))
                    scene.errors.extend(batch.errors)
                    self.errors.extend(batch.errors)
                    if self.stop_on_error:
                        return

                if self.max_lines and self.lines_processed >= self.max_lines:
                    self._emit_info(_("Reached max_lines limit of ({lines} lines)... finishing").format(lines=self.max_lines))
                    break

            # Update the scene summary based on the best available information (we hope)
            scene.summary = self._get_best_summary([scene.summary, context.get('scene'), context.get('summary')])

            # Notify observers the scene was translated
            self.events.scene_translated.send(self, scene=scene)

        except (TranslationAbortedError, TranslationImpossibleError) as e:
            raise

    def TranslateBatch(self, batch : SubtitleBatch, line_numbers : list[int]|None, context : dict[str,Any]|None):
        """
        Send batches of subtitles for translation, building up context.
        """
        if self.aborted:
            return

        if self.reparse and batch.translation:
            self._emit_info(_("Reparsing scene {scene} batch {batch} with {count} lines...").format(scene=batch.scene, batch=batch.number, count=len(batch.originals)))
            self.ProcessBatchTranslation(batch, batch.translation, line_numbers)
            return

        if self.resume and not self.retranslate and batch.all_translated:
            self._emit_info(_("Scene {scene} batch {batch} already translated {lines} lines...").format(scene=batch.scene, batch=batch.number, lines=batch.size))
            return

        originals, context = self.PreprocessBatch(batch, context)

        logging.debug(f"Translating scene {batch.scene} batch {batch.number} with {len(originals)} lines...")

        # Build summaries context
        context['batch'] = f"Scene {batch.scene} batch {batch.number}"
        if batch.summary:
            context['summary'] = batch.summary

        instructions = self.instructions.instructions
        if not instructions:
            raise TranslationImpossibleError(_("No instructions provided for translation"))

        batch.prompt = self.client.BuildTranslationPrompt(self.user_prompt, instructions, originals, context)

        if self.preview:
            return

        # Ask the client to do the translation
        streaming_callback = self._create_streaming_callback(batch, line_numbers) if self.client.enable_streaming else None
        translation : Translation|None = self.client.RequestTranslation(batch.prompt, streaming_callback=streaming_callback)

        if (translation and translation.reached_token_limit) and not self.aborted:
            # Try again without the context to keep the tokens down
            # TODO: better to split the batch into smaller chunks
            logging.warning(_("Hit API token limit, retrying batch without context..."))
            batch.prompt.GenerateMessages(instructions, batch.originals, {})

            translation = self.client.RequestTranslation(batch.prompt, streaming_callback=streaming_callback)

        if not self.aborted:
            if not translation:
                raise TranslationError(f"Unable to translate scene {batch.scene} batch {batch.number}")

            # Process the response
            self.ProcessBatchTranslation(batch, translation, line_numbers)

            # Consider retrying if there were errors
            if batch.errors and self.retry_on_error:
                logging.warning(_("Scene {scene} batch {batch} failed validation, requesting retranslation").format(scene=batch.scene, batch=batch.number))
                self.RequestRetranslation(batch, line_numbers=line_numbers, context=context)

            # Update the context, unless it's a retranslation pass
            if not self.retranslate and not self.aborted:
                context['summary'] = self._get_best_summary([translation.summary, batch.summary])
                context['scene'] = self._get_best_summary([translation.scene, context.get('scene')])
                context['synopsis'] = translation.synopsis or context.get('synopsis', "")
                #context['names'] = translation.names or context.get('names', []) or options.get('names')
                batch.UpdateContext(context)

    def PreprocessBatch(self, batch : SubtitleBatch, context : dict[str,Any]|None = None) -> tuple[list[SubtitleLine], dict[str, Any]]:
        """
        Preprocess the batch before translation
        """
        context = context or {}
        if batch.context and (self.retranslate or self.reparse):
            # If it's a retranslation, restore context from the batch
            for key, value in batch.context.items():
                context[key] = value

        # Apply any substitutions to the input
        replacements = batch.PerformInputSubstitutions(self.substitutions)

        if replacements:
            replaced : list[str] = [f"{Linearise(k)} -> {Linearise(v)}" for k,v in replacements.items()]
            self._emit_info(_("Made substitutions in input:\n{replaced}").format(replaced=linesep.join(replaced)))
            batch.AddContext('replacements', replaced)

        # Filter out empty lines
        originals = [ line for line in batch.originals if line.text and line.text.strip() ]

        # Apply the max_lines limit
        with self.lock:
            line_count = min(self.max_lines - self.lines_processed, len(originals)) if self.max_lines else len(originals)
            self.lines_processed += line_count
            if len(originals) > line_count:
                self._emit_info(_("Truncating batch to remain within max_lines"))
                originals = originals[:line_count] if line_count > 0 else []

        return originals, context

    def ProcessBatchTranslation(self, batch : SubtitleBatch, translation : Translation, line_numbers : list[int]|None):
        """
        Attempt to extract translation from the API response
        """
        if not translation:
            raise NoTranslationError("No translation provided")

        if not translation.has_translation:
            raise TranslationError("Translation contains no translated text", translation=translation)

        logging.debug(f"Scene {batch.scene} batch {batch.number} translation:\n{translation.text}\n")

        # Apply the translation to the subtitles
        parser : TranslationParser = self.client.GetParser(self.task_type)

        parser.ProcessTranslation(translation)

        # Try to match the translations with the original lines
        translated, unmatched = parser.MatchTranslations(batch.originals)

        # Assign the translated lines to the batch
        if line_numbers:
            translated = [line for line in translated if line.number in line_numbers]

        batch._translated = MergeTranslations(batch.translated or [], translated)

        batch.translation = translation
        batch.errors = [err for err in parser.errors if isinstance(err, str) or isinstance(err, SubtitleError)]

        # Emit any warnings from the parser
        for warning in parser.warnings:
            self._emit_warning(warning)

        if batch.untranslated and not self.max_lines:
            self._emit_warning(_("Unable to match {count} lines with a source line").format(count=len(unmatched)))
            batch.AddContext('untranslated_lines', [f"{item.number}. {item.text}" for item in batch.untranslated])

        # Apply any word/phrase substitutions to the translation
        replacements = batch.PerformOutputSubstitutions(self.substitutions)

        if replacements:
            replaced = [f"{k} -> {v}" for k,v in replacements.items()]
            self._emit_info(_("Made substitutions in output:\n{replaced}").format(replaced=linesep.join(replaced)))

        # Perform substitutions on the output
        translation.PerformSubstitutions(self.substitutions)

        # Post-process the translation
        if self.postprocessor:
            batch._translated = self.postprocessor.PostprocessSubtitles(batch.translated)

            self._emit_info(_("Scene {scene} batch {batch}: {translated} lines and {untranslated} untranslated.").format(
                scene=batch.scene, 
                batch=batch.number, 
                translated=len(batch.translated or []), 
                untranslated=len(batch.untranslated or []))
                )

        if translation.summary and translation.summary.strip():
            self._emit_info(_("Summary: {summary}").format(summary=translation.summary))

    def RequestRetranslation(self, batch : SubtitleBatch, line_numbers : list[int]|None = None, context : dict[str, str]|None = None):
        """
        Ask the client to retranslate the input and correct errors
        """
        translation : Translation|None = batch.translation
        if not translation:
            raise TranslationError("No translation to retranslate")

        prompt : TranslationPrompt|None = batch.prompt
        if not prompt or not prompt.messages:
            raise TranslationError("No prompt to retranslate")

        if not self.instructions.retry_instructions:
            raise TranslationError("No retry instructions provided")

        if not translation.text:
            raise TranslationError("No translation text to retranslate", translation=translation)

        retry_instructions = self.instructions.retry_instructions
        if retry_instructions is None:
            return

        prompt.GenerateRetryPrompt(translation.text, retry_instructions, batch.errors)

        # Let's raise the temperature a little bit
        temperature = self.client.temperature or 0.0
        retry_temperature = min(temperature + 0.1, 1.0)

        retranslation : Translation|None = self.client.RequestTranslation(prompt, retry_temperature)

        if self.aborted:
            return None

        if not isinstance(retranslation, Translation):
            raise TranslationError("Retranslation is not the expected type", translation=retranslation)

        logging.debug(f"Scene {batch.scene} batch {batch.number} retranslation:\n{retranslation.text}\n")

        self.ProcessBatchTranslation(batch, retranslation, line_numbers)

        if batch.errors:
            self._emit_warning(_("Retry failed validation: {errors}").format(errors=FormatErrorMessages(batch.errors)))
        else:
            self._emit_info(_("Retry passed validation"))

    def _get_best_summary(self, candidates : list[str|None]) -> str|None:
        """
        Generate a summary of the translated subtitles
        """
        movie_name : str|None = self.settings.get_str( 'movie_name', None)
        movie_name = str(movie_name).strip() if movie_name else None
        max_length : int|None = self.max_summary_length

        for candidate in candidates:
            if candidate is None:
                continue

            sanitised = SanitiseSummary(candidate, movie_name, max_length)
            if sanitised:
                if len(sanitised) < len(candidate):
                    self._emit_info(_("Summary was truncated from {original} to {truncated} characters").format(original=len(candidate), truncated=len(sanitised)))
                return sanitised

        return None

    def _create_streaming_callback(self, batch : SubtitleBatch, line_numbers : list[int]|None) -> StreamingCallback:
        """
        Create a streaming callback that processes partial translations and emits batch_updated events
        """
        def streaming_callback(partial_translation : Translation):
            if self.aborted or not partial_translation:
                return

            try:
                # Process the partial translation (without validation)
                self._process_partial_translation(batch, partial_translation, line_numbers)

                # Emit batch_updated event with the updated batch
                self.events.batch_updated.send(self, batch=batch)

            except Exception as e:
                logging.warning(f"Error processing streaming update for scene {batch.scene} batch {batch.number}: {e}")

        return streaming_callback

    def _process_partial_translation(self, batch : SubtitleBatch, translation : Translation, line_numbers : list[int]|None):
        """
        Process a partial translation without validation (streaming updates only)
        """
        if not translation or not translation.has_translation:
            return

        parser = self.client.GetParser(self.task_type)
        try:
            parser.ProcessTranslation(translation, validate=False)

            translated, _ = parser.MatchTranslations(batch.originals)

            if line_numbers:
                translated = [line for line in translated if line.number in line_numbers]

            # Merge with existing translations (MergeTranslations is already imported at top)
            # Todo: we should use a SubtitleEditor to merge changes 
            batch.translated = MergeTranslations(batch.translated or [], translated)

            # Note: We don't set errors for partial translations to avoid false validation failures

        except Exception:
            pass

    def _emit_error(self, message : str):
        """Emit an error event"""
        self.events.error.send(self, message=message)

    def _emit_warning(self, message : str):
        """Emit a warning event"""
        self.events.warning.send(self, message=message)

    def _emit_info(self, message : str):
        """Emit an info event"""
        self.events.info.send(self, message=message)
